# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2020 Andrew Rechnitzer
# Copyright (C) 2018 Elvis Cai
# Copyright (C) 2019-2020 Colin B. Macdonald
# Copyright (C) 2020 Victoria Schuster

"""
The Plom Marker client
"""

__copyright__ = "Copyright (C) 2018-2020 Andrew Rechnitzer and others"
__credits__ = ["Andrew Rechnitzer", "Elvis Cai", "Colin Macdonald", "Victoria Schuster"]
__license__ = "AGPL-3.0-or-later"


import json
import logging
import os
import secrets

# in order to get shortcuts under OSX this needs to set this.... but only osx.
# To test platform
import platform
import queue
import shutil
import tempfile
import threading
import time
from collections import defaultdict

from PyQt5.QtCore import (
    Qt,
    QSortFilterProxyModel,
    QTimer,
    QThread,
    pyqtSlot,
    pyqtSignal,
)
from PyQt5.QtGui import QStandardItem, QStandardItemModel
from PyQt5.QtWidgets import (
    QDialog,
    QMessageBox,
    QProgressDialog,
    QWidget,
)

from plom.plom_exceptions import *
from .annotator import Annotator
from .comment_list import AddTagBox, commentLoadAll, commentIsVisible
from .examviewwindow import ExamViewWindow
from .origscanviewer import GroupView, SelectTestQuestion
from .uiFiles.ui_marker import Ui_MarkerWindow
from .useful_classes import ErrorMessage, SimpleMessage

if platform.system() == "Darwin":
    from PyQt5.QtGui import qt_set_sequence_auto_mnemonic

    qt_set_sequence_auto_mnemonic(True)

log = logging.getLogger("marker")

# set up variables to store paths for marker and id clients
tempDirectory = tempfile.TemporaryDirectory(prefix="plom_")
directoryPath = tempDirectory.name


# Read https://mayaposch.wordpress.com/2011/11/01/how-to-really-truly-use-qthreads-the-full-explanation/
# and https://stackoverflow.com/questions/6783194/background-thread-with-qthread-in-pyqt
# and finally https://woboq.com/blog/qthread-you-were-not-doing-so-wrong.html
# I'll do it the simpler subclassing way
class BackgroundDownloader(QThread):
    """
    Downloads exams in background.

    Notes:
        Read https://mayaposch.wordpress.com/2011/11/01/how-to-really-truly-use-qthreads-the-full-explanation/
        and https://stackoverflow.com/questions/6783194/background-thread-with-qthread-in-pyqt
        and finally https://woboq.com/blog/qthread-you-were-not-doing-so-wrong.html
        (Done in the simpler subclassing way.)

    """

    downloadSuccess = pyqtSignal(
        str, list, list, str, str
    )  # [task, files, image_md5s, tags, integrity_check]
    downloadNoneAvailable = pyqtSignal()
    downloadFail = pyqtSignal(str)

    def __init__(self, question, version):
        """
        Initializes a new downloader.

        Args:
            question (str): question number
            version (str): version number.

        Notes:
            question/version may be able to be type int as well.

        """
        QThread.__init__(self)
        self.question = question
        self.version = version
        self.workingDirectory = directoryPath

    def run(self):
        """
        Runs the background downloader.

        Notes:
            Overrides run method of QThread.

        Returns:
            None

        """
        attempts = 0
        while True:
            attempts += 1
            # little sanity check - shouldn't be needed.
            # TODO remove.
            if attempts > 5:
                return
            # ask server for task-code of next task
            try:
                task = messenger.MaskNextTask(self.question, self.version)
                if not task:  # no more tests left
                    self.downloadNoneAvailable.emit()
                    self.quit()
                    return
            except PlomSeriousException as err:
                self.downloadFail.emit(str(err))
                self.quit()
                return

            try:
                image_metadata, tags, integrity_check = messenger.MclaimThisTask(task)
                break
            except PlomTakenException as err:
                log.info("will keep trying as task already taken: {}".format(err))
                continue
            except PlomSeriousException as err:
                self.downloadFail.emit(str(err))
                self.quit()

        image_md5s = [x[1] for x in image_metadata]
        # Image names = "<task>.<imagenumber>.<extension>"
        image_fnames = []
        for i, row in enumerate(image_metadata):
            # try-except? how does this fail?
            im_bytes = messenger.MrequestOneImage(task, row[0], row[1])
            tmp = os.path.join(self.workingDirectory, "{}.{}.image".format(task, i))
            image_fnames.append(tmp)
            with open(tmp, "wb+") as fh:
                fh.write(im_bytes)
        self.downloadSuccess.emit(task, image_fnames, image_md5s, tags, integrity_check)
        self.quit()


class BackgroundUploader(QThread):
    """Uploads exams in Background."""

    uploadSuccess = pyqtSignal(str, int, int)
    uploadKnownFail = pyqtSignal(str, str)
    uploadUnknownFail = pyqtSignal(str, str)

    def enqueueNewUpload(self, *args):
        """
        Places something in the upload queue.

        Note:
            If you call this from the main thread, this code runs in the
            main thread. That is ok because Queue is threadsafe, but it's
            important to be aware, not all code in this object runs in the new
            thread: it depends on where that code is called!

        Args:
            *args: all input arguments are cached and will eventually be
                passed untouched to the `upload` function.  There is one
                exception: `args[0]` is assumed to contain the task str
                of the form `"q1234g9"` for printing debugging messages.

        Returns:
            None
        """
        log.debug("upQ enqueing item from main thread " + str(threading.get_ident()))
        self.q.put(args)

    def isEmpty(self):
        """
        Checks if the upload queue is empty.

        Returns:
            True if the upload queue is empty, false otherwise.

        """
        return self.q.empty()

    def run(self):
        """
        Runs the uploader in background.

        Notes:
            Overrides run method of Qthread.

        Returns:
            None

        """

        def tryToUpload():
            # define this inside run so it will run in the new thread
            # https://stackoverflow.com/questions/52036021/qtimer-on-a-qthread
            from queue import Empty as EmptyQueueException

            try:
                data = self.q.get_nowait()
            except EmptyQueueException:
                return
            code = data[0]  # TODO: remove so that queue needs no knowledge of args
            log.info("upQ thread: popped code {} from queue, uploading".format(code))
            upload(
                *data,
                knownFailCallback=self.uploadKnownFail.emit,
                unknownFailCallback=self.uploadUnknownFail.emit,
                successCallback=self.uploadSuccess.emit,
            )

        self.q = queue.Queue()
        log.info("upQ thread: starting with new empty queue and starting timer")
        # TODO: Probably don't need the timer: after each enqueue, signal the
        # QThread (in the new thread's event loop) to call tryToUpload.
        timer = QTimer()
        timer.timeout.connect(tryToUpload)
        timer.start(250)
        self.exec_()


def upload(
    task,
    grade,
    filenames,
    mtime,
    question,
    ver,
    tags,
    integrity_check,
    image_md5_list,
    knownFailCallback=None,
    unknownFailCallback=None,
    successCallback=None,
):
    """
    Uploads a paper.

    Args:
        task (str): the Task ID for the page being uploaded. Takes the form
            "q1234g9" = test 1234 question 9.
        grade (int): grade given to question.
        filenames (list[str]): a list containing the annotated file's name,
            the .plom file's name and the comment file's name, in that order.
        mtime (int): the marking time (s) for this specific question.
        question (int or str): the question number
        ver (int or str): the version number
        tags (str): any tags associated with this exam.
        integrity_check (str): the integrity_check string of the task.
        image_md5_list (list[str]): a list of image md5sums used.
        knownFailCallback: if we fail in a way that is reasonably expected,
            call this function.
        unknownFailCallback: if we fail but don't really know why or what
            do to, call this function.
        successCallback: a function to call when we succeed.

    Returns:
        None

    Raises:
        PlomSeriousException if elements in filenames do not correspond to
            the same exam.

    """
    # do name sanity checks here
    aname, pname, cname = filenames

    if not (
        task.startswith("q")
        and os.path.basename(aname) == "G{}.png".format(task[1:])
        and os.path.basename(pname) == "G{}.plom".format(task[1:])
        and os.path.basename(cname) == "G{}.json".format(task[1:])
    ):
        raise PlomSeriousException(
            "Upload file names mismatch [{}, {}, {}] - this should not happen".format(
                aname, pname, cname
            )
        )
    try:
        msg = messenger.MreturnMarkedTask(
            task,
            question,
            ver,
            grade,
            mtime,
            tags,
            aname,
            pname,
            cname,
            integrity_check,
            image_md5_list,
        )
    except (PlomTaskChangedError, PlomTaskDeletedError, PlomConflict) as ex:
        knownFailCallback(task, str(ex))
        # probably previous call does not return: it forces a crash
        return
    except PlomException as ex:
        unknownFailCallback(task, str(ex))
        return

    numDone = msg[0]
    numTotal = msg[1]
    successCallback(task, numDone, numTotal)


class ExamQuestion:
    """
    A class storing identifying information for one Exam Question.

    A simple container for storing a groupimage's task ID,
    numer, group, version, status, the mark, the original image
    filename, the annotated image filename, the mark, and the
    time spent marking the groupimage.
    """

    def __init__(
        self,
        task,
        fnames=[],
        stat="untouched",
        mrk="-1",
        mtime="0",
        tags="",
        integrity_check="",
        image_md5s=[],
    ):
        """
        Initializes an exam question.

        Args:
            task (str): the Task ID for the page being uploaded. Takes the form
            "q1234g9" = test 1234 question 9.
            fnames (list[str]): a list containing the filename for the
            original image/images for the test question.
            stat (str): test status.
            mrk (int): the mark of the question.
            mtime (int): marking time spent on that page in seconds.
            tags (str): Tags corresponding to the exam.
            integrity_check (str): integrity_check = concat of md5sums of underlying images
            image_md5s (list[str]): a list containing the md5sums of the images for the test question.

        Notes:
            By default set mark to be negative (since 0 is a possible mark)
        """
        self.prefix = task
        self.status = stat
        self.mark = mrk
        self.originalFiles = fnames
        self.image_md5s = image_md5s
        self.annotatedFile = ""  # The filename for the (future) annotated image
        self.plomFile = ""  # The filename for the (future) plom file
        self.markingTime = mtime
        self.tags = tags
        self.integrity_check = integrity_check


class MarkerExamModel(QStandardItemModel):
    """A tablemodel for handling the group image marking data."""

    def __init__(self, parent=None):
        """
        Initializes a new MarkerExamModel.

        Args:
            parent (QStandardItemModel): MarkerExamModel's Parent.

        """
        QStandardItemModel.__init__(self, parent)
        self.setHorizontalHeaderLabels(
            [
                "Task",
                "Status",
                "Mark",
                "Time (s)",
                "Tag",
                "OriginalFiles",
                "AnnotatedFile",
                "PlomFile",
                "PaperDir",
                "integrity_check",
                "image_md5s",
            ]
        )

    def addPaper(self, paper):
        """
        Adds a paper to self.

        Args:
            paper (ExamQuestion): the paper to be added

        Returns:
            r (int): the row identifier of the added paper.

        """
        # check if paper is already in model - if so, delete it and add it back with the new data.
        # **but** be careful - if annotation in progress then ??
        try:
            r = self._findTask(paper.prefix)
        except ValueError as err:
            pass
        else:
            ErrorMessage(
                "Task {} has been modified by server - you will need to annotate it again.".format(
                    paper.prefix
                )
            ).exec_()
            self.removeRow(r)
        # Append new groupimage to list and append new row to table.
        r = self.rowCount()
        self.appendRow(
            [
                QStandardItem(paper.prefix),
                QStandardItem(paper.status),
                QStandardItem(str(paper.mark)),
                QStandardItem(str(paper.markingTime)),
                QStandardItem(paper.tags),
                # TODO - work out how to store list more directly rather than as a string-rep of the list of file names - maybe json? similarly for image_md5s
                QStandardItem(repr(paper.originalFiles)),
                QStandardItem(paper.annotatedFile),
                QStandardItem(paper.plomFile),
                QStandardItem("placeholder"),
                # todo - reorder these?
                QStandardItem(paper.integrity_check),
                QStandardItem(json.dumps(paper.image_md5s)),  # json (list[str] -> str)
            ]
        )
        return r

    def _getPrefix(self, r):
        """
        Return the prefix of the image

        Args:
            r (int): the row identifier of the paper.

        Returns:
            (str): the prefix of the image

        """
        return self.data(self.index(r, 0))

    def _getStatus(self, r):
        """
        Returns the status of the image.

        Args:
            r (int): the row identifier of the paper.

        Returns:
            (str): the status of the image

        """
        return self.data(self.index(r, 1))

    def _setStatus(self, r, stat):
        """
        Sets the status of the image.

        Args:
            r (int): the row identifier of the paper.
            stat (str): the new status of the image.

        Returns:
            None

        """
        self.setData(self.index(r, 1), stat)

    def _getAnnotatedFile(self, r):
        """
        Returns the filename of the annotated image.

        Args:
            r (int): the row identifier of the paper.

        Returns:
            (str): the filename of the annotated image

        """
        return self.data(self.index(r, 6))

    def _setAnnotatedFile(self, r, aname, pname):
        """
        Set the file name for the annotated image.

        Args:
            r (int): the row identifier of the paper.
            aname (str): the name for the annotated file.
            pname (str): the name for the .plom file

        Returns:
            None

        """
        self.setData(self.index(r, 6), aname)
        self.setData(self.index(r, 7), pname)

    def _setPaperDir(self, r, tdir):
        """
        Sets the paper directory for the given paper.
        Args:
            r (int): the row identifier of the paper.
            tdir (dir): a temporary directory for this paper.

        Returns:
            None
        """
        self.setData(self.index(r, 8), tdir)

    def _clearPaperDir(self, r):
        """
        Clears the paper directory for the given paper.

        Args:
            r (int): the row identifier of the paper.

        Returns:
            None
        """
        self._setPaperDir(r, None)

    def _getPaperDir(self, r):
        """
        Returns the paper directory for the given paper.

        Args:
            r (int): the row identifier of the paper.

        Returns:
            (dir): a temporary directory for this paper.
        """
        return self.data(self.index(r, 8))

    def _findTask(self, task):
        """
        Return the row index of this task.

        Args:
            task (str): the task for the image files to be loaded from.
                Takes the form "q1234g9" = test 1234 question 9

        Raises:
             ValueError if not found.
        """
        r0 = []
        for r in range(self.rowCount()):
            if self._getPrefix(r) == task:
                r0.append(r)

        if len(r0) == 0:
            raise ValueError("task {} not found!".format(task))
        elif not len(r0) == 1:
            raise ValueError(
                "Repeated task {} in rows {}  This should not happen!".format(task, r0)
            )
        return r0[0]

    def _setDataByTask(self, task, n, stuff):
        """
        Find the row identifier with `task` and sets `n`th column to `stuff`.

        Args:
            task (str): the task for the image files to be loaded from.
            n (int): the column to be loaded into.
            stuff: whatever is being added.

        Returns:
            None
        """
        r = self._findTask(task)
        self.setData(self.index(r, n), stuff)

    def _getDataByTask(self, task, n):
        """
        Returns contents of task in `n`th column.

        Args:
            task (str): the task for the image files to be loaded from.
            n (int): the column to return from.

        Returns:
            Contents of task in `n`th column.
        """
        r = self._findTask(task)
        return self.data(self.index(r, n))

    def getStatusByTask(self, task):
        """Return status for task (task(str) defined above.)"""
        return self._getDataByTask(task, 1)

    def setStatusByTask(self, task, st):
        """Set status for task, ((task(str) and st(str) defined above.)"""
        self._setDataByTask(task, 1, st)

    def getTagsByTask(self, task):
        """Return tags for task, (task(str) defined above.)"""
        return self._getDataByTask(task, 4)

    def setTagsByTask(self, task, tags):
        """Set tags for task, (task (str) defined above.)"""
        return self._setDataByTask(task, 4, tags)

    def getAllTags(self):
        """Return all tags as a set."""
        tags = set()
        for r in range(self.rowCount()):
            v = self.data(self.index(r, 4))
            if len(v) > 0:
                tags.add(v)
        return tags

    def getMTimeByTask(self, task):
        """Return total marking time (s) for task, (task(str), return (int).)"""
        return int(self._getDataByTask(task, 3))

    def getPaperDirByTask(self, task):
        """Return temporary directory for this task, (task(str) defined above)"""
        return self._getDataByTask(task, 8)

    def setPaperDirByTask(self, task, tdir):
        """
        Set temporary directory for this grading.

        Args:
            task (str): the task for the image files to be loaded from.
            tdir (dir): the temporary directory for task to be set to.

        Returns:
            None
        """
        self._setDataByTask(task, 8, tdir)

    def getOriginalFiles(self, task):
        """Return filename for original un-annotated image as string."""
        return eval(self._getDataByTask(task, 5))

    def _setOriginalFiles(self, task, fnames):
        """Set the original un-annotated image filenames."""
        self._setDataByTask(task, 5, repr(fnames))

    def _setImageMD5s(self, task, image_md5s):
        """Set the md5sums of the original image files."""
        self._setDataByTask(task, 10, json.dumps(image_md5s))

    def setOriginalFilesAndMD5s(self, task, fnames, image_md5s):
        """Set the original un-annotated image filenames and their md5sums."""
        # TODO: better to store these in zipped pairs!
        self._setOriginalFiles(task, fnames)
        self._setImageMD5s(task, image_md5s)

    def setAnnotatedFile(self, task, aname, pname):
        """Set the annotated image and .plom file names."""
        self._setDataByTask(task, 6, aname)
        self._setDataByTask(task, 7, pname)

    def getIntegrityCheck(self, task):
        """Return integrity_check for task as string."""
        return self._getDataByTask(task, 9)

    def getImageMD5List(self, task):
        """Return image_md5sum list for task as a string."""
        return self._getDataByTask(task, 10)

    def markPaperByTask(self, task, mrk, aname, pname, mtime, tdir):
        """
        Add marking data for the given task.

        Args:
            task (str): the task for the image files to be loaded from.
            mrk (int): the mark for this paper.
            aname (str): the annotated file name.
            pname (str): the .plom file name.
            mtime (int): total marking time in seconds.
            tdir (dir): the temporary directory for task to be set to.

        Returns:
            None

        """
        # There should be exactly one row with this Task
        r = self._findTask(task)
        # When marked, set the annotated filename, the plomfile, the mark,
        # and the total marking time (in case it was annotated earlier)
        mt = int(self.data(self.index(r, 3)))
        # total elapsed time.
        self.setData(self.index(r, 3), mtime + mt)
        self._setStatus(r, "uploading...")
        self.setData(self.index(r, 2), mrk)
        self._setAnnotatedFile(r, aname, pname)
        self._setPaperDir(r, tdir)

    def deferPaper(self, task):
        """ Sets the status for the task's paper to deferred. """
        self.setStatusByTask(task, "deferred")

    def removePaper(self, task):
        """ Removes the task's paper from self."""
        r = self._findTask(task)
        self.removeRow(r)

    def countReadyToMark(self):
        """Returns the number of untouched Papers."""
        count = 0
        for r in range(self.rowCount()):
            if self._getStatus(r) == "untouched":
                count += 1
        return count


##########################
class ProxyModel(QSortFilterProxyModel):
    """A proxymodel wrapper to handle filtering and sorting of table model."""

    def __init__(self, parent=None):
        """
        Initializes a new ProxyModel object.

        Args:
            parent (QObject): self's parent.
        """
        QSortFilterProxyModel.__init__(self, parent)
        self.setFilterKeyColumn(4)
        self.filterString = ""
        self.invert = False

    def lessThan(self, left, right):
        """
        Sees if left data is less than right data.

        Args:
            left (QModelIndex): the left map.
            right (QModelIndex):

        Returns:
            True if left data is less than right data.

        """
        # Check to see if data is integer, and compare that
        try:
            lv = int(left.data())
            rv = int(right.data())
            return lv < rv
        except ValueError:
            # else let qt handle it.
            return left.data() < right.data()

    def setFilterString(self, flt):
        """
        Sets the Filter String.

        Args:
            flt (str): the string to filter by

        Returns:
            None

        """
        self.filterString = flt

    def filterTags(self, invert=False):
        """
        Sets the Filter Tags based on string.

        Args:
            invert (bool): True if looking for files that do not have given
                filter string, false otherwise.

        Returns:
            None

        """
        self.invert = invert
        self.setFilterFixedString(self.filterString)

    def filterAcceptsRow(self, pos, index):
        """
        Checks if a row fits the given filter.

        Notes:
            Overrides base method.

        Args:
            pos (int): row desired.
            index (any): unused.

        Returns:
            True if filter accepts the row, False otherwise.

        """
        if (len(self.filterString) == 0) or (
            self.filterString.casefold()
            in self.sourceModel().data(self.sourceModel().index(pos, 4)).casefold()
        ):
            # we'd return true here, unless INVERT, then false
            if self.invert:
                return False
            else:
                return True
        else:  # we'd return false here, unless invert, then true
            if self.invert:
                return True
            else:
                return False

    def getPrefix(self, r):
        """
        Returns the prefix of inputted row index.

        Args:
            r (int): the row identifier of the paper.

        Returns:
            (str): the prefix of the paper indicated by r.

        """
        return self.data(self.index(r, 0))

    def getStatus(self, r):
        """
        Returns the status of inputted row index.

        Args:
            r (int): the row identifier of the paper.

        Returns:
            (str): the status of the paper indicated by r.

        """
        # Return the status of the image
        return self.data(self.index(r, 1))

    def getOriginalFiles(self, r):
        """
        Returns the file names of the un-annotated image of r.

        Args:
            r (int): the row identifier of the paper.

        Returns:
            (str): the file name of the original, un-annotated image of the
                paper in r.

        """
        return eval(self.data(self.index(r, 5)))

    def getAnnotatedFile(self, r):
        """
        Returns the file names of the annotated image of r.

        Args:
            r (int): the row identifier of the paper.

        Returns:
            (str): the file name of the annotated image of the paper in r.

        """
        return self.data(self.index(r, 6))

    def rowFromTask(self, task):
        """Return the row index (int) of this task (str) or None if absent."""
        r0 = []
        for r in range(self.rowCount()):
            if self.getPrefix(r) == task:
                r0.append(r)

        if len(r0) == 0:
            return None
        elif not len(r0) == 1:
            raise ValueError(
                "Repeated task {} in rows {}  This should not happen!".format(task, r0)
            )
        return r0[0]


class MarkerClient(QWidget):
    """
    Setup for marking client and annotator

    Notes:
        TODO: should be a QMainWindow but at any rate not a Dialog
        TODO: should this be parented by the QApplication?
    """

    my_shutdown_signal = pyqtSignal(int, list)

    def __init__(self, Qapp):
        """
        Initialize a new MarkerClient

        Args:
            Qapp(QApplication): Main client application

        """
        super(MarkerClient, self).__init__()
        self.Qapp = Qapp

        # instance vars we can initialize now
        self.workingDirectory = (
            directoryPath
            # local temp directory for image files and the class list.
        )
        self.viewFiles = []  # For viewing the whole paper we'll need these two lists.
        self.maxMark = -1  # temp value
        self.examModel = (
            MarkerExamModel()
        )  # Exam model for the table of groupimages - connect to table
        self.prxM = ProxyModel()  # set proxy for filtering and sorting
        self.testImg = (
            ExamViewWindow()
        )  # A view window for the papers so user can zoom in as needed.
        self.annotatorSettings = defaultdict(
            lambda: None
        )  # settings variable for annotator settings (initially None)
        self.commentCache = {}  # cache for Latex Comments
        self.backgroundDownloader = None
        self.backgroundUploader = None

        self.allowBackgroundOps = True
        self.canViewAll = False

        # instance vars that get initialized later
        self.question = None
        self.version = None
        self.exam_spec = None
        self.ui = None
        self.canViewAll = None

    def setup(self, markerMessenger, question, version, lastTime):
        """
        Performs setup procedure for markerClient.

        TODO: Change the __init__ params to include the
            params below and move this method into init

        TODO: verify all lastTime Params, there are almost certainly some missing

        Args:
            markerMessenger (Messenger): messenger client for communicating with server
            question (int): question number.
            version (int): version number
            lastTime (dict): a dictionary containing
                 {"user": username
                "server": serverNumber
                 "question": question number
                 "version": version number
                 "fontsize"
                 "POWERUSER"
                 "FOREGROUND"
                 "upDown": marking style (up vs down)
                 "LogLevel"
                 "LogToFile"
                 "CommentsWarnings"
                 "MarkWarnings"
                 "mouse": left or right mouse hand
                 "SidebarOnRight": True if sidebar is on right
                  }
                and potentially others

        Returns:
            None

        """

        global messenger
        messenger = markerMessenger  # initializes global messenger
        self.question = question
        self.version = version

        # Get the number of Tests, Pages, Questions and Versions
        try:
            self.exam_spec = messenger.get_spec()
        except PlomSeriousException as err:
            self.throwSeriousError(err, rethrow=False)
            return

        self.UIInitialization()
        self.applyLastTimeOptions(lastTime)
        self.connectGuiButtons()
        self.setMarkStyleID()

        if not self.getMaxMark():  # indicates exception was caught
            return
        self.ui.maxscoreLabel.setText(str(self.maxMark))

        try:
            self.loadMarkedList()  # Get list of papers already marked and add to table.
        except PlomSeriousException as err:
            self.throwSeriousError(err)
            return

        # Keep the original format around in case we need to change it
        self.ui._cachedProgressFormatStr = self.ui.mProgressBar.format()
        self.updateProgress()  # Update counts

        # Connect the view **after** list updated.
        # Connect the table-model's selection change to appropriate function
        self.ui.tableView.selectionModel().selectionChanged.connect(self.updateImg)

        self.requestNext()  # Get a question to mark from the server
        self.testImg.resetB.animateClick()  # reset the view so whole exam shown.
        # resize the table too.
        QTimer.singleShot(100, self.ui.tableView.resizeRowsToContents)
        log.debug("Marker main thread: " + str(threading.get_ident()))

        if self.allowBackgroundOps:
            self.backgroundUploader = BackgroundUploader()
            self.backgroundUploader.uploadSuccess.connect(self.backgroundUploadFinished)
            self.backgroundUploader.uploadKnownFail.connect(
                self.backgroundUploadFailedServerChanged
            )
            self.backgroundUploader.uploadUnknownFail.connect(
                self.backgroundUploadFailed
            )
            self.backgroundUploader.start()
        self.cacheLatexComments()  # Now cache latex for comments:

    def applyLastTimeOptions(self, lastTime):
        """
        Applies all settings from previous client.

        Args:
            lastTime (dict): a dictionary containing information
                            for prev client. See setup for more info.

        Returns:
            None
        """
        self.annotatorSettings["commentWarnings"] = lastTime.get("CommentWarnings")
        self.annotatorSettings["markWarnings"] = lastTime.get("MarkWarnings")

        if lastTime.get("POWERUSER", False):
            # if POWERUSER is set, disable warnings and allow viewing all
            self.canViewAll = True

        if lastTime.get("FOREGROUND", False):
            self.allowBackgroundOps = False

        self.ui.sidebarRightCB.setChecked(lastTime.get("SidebarOnRight", False))

        if lastTime["upDown"] == "up":
            self.ui.markUpRB.animateClick()

        elif lastTime["upDown"] == "down":
            self.ui.markDownRB.animateClick()

        if lastTime["mouse"] == "left":
            self.ui.leftMouseCB.setChecked(True)

    def UIInitialization(self):
        """
        Startup procedure for the user interface

        Returns:
            None: Modifies self.ui

        """
        # Fire up the user interface
        self.ui = Ui_MarkerWindow()
        self.ui.setupUi(self)
        self.setWindowTitle('Plom Marker: "{}"'.format(self.exam_spec["name"]))
        # Paste the username, question and version into GUI.
        self.ui.userLabel.setText(messenger.whoami())
        self.ui.infoBox.setTitle(
            "Marking Q{} (ver. {}) of “{}”".format(
                self.question, self.version, self.exam_spec["name"]
            )
        )

        self.prxM.setSourceModel(self.examModel)
        self.ui.tableView.setModel(self.prxM)
        # hide various columns without end-user useful info
        self.ui.tableView.hideColumn(5)
        self.ui.tableView.hideColumn(6)
        self.ui.tableView.hideColumn(7)
        self.ui.tableView.hideColumn(8)
        self.ui.tableView.hideColumn(9)
        self.ui.tableView.hideColumn(10)

        # Double-click or signal fires up the annotator window
        self.ui.tableView.doubleClicked.connect(self.annotateTest)
        self.ui.tableView.annotateSignal.connect(self.annotateTest)
        # A view window for the papers so user can zoom in as needed.
        # Paste into appropriate location in gui.
        self.ui.gridLayout_6.addWidget(self.testImg, 0, 0)

    def connectGuiButtons(self):
        """
        Connect gui buttons to appropriate functions

        Notes:
            TODO: remove the total-radiobutton

        Returns:
            None - Modifies self.ui
        """
        self.ui.closeButton.clicked.connect(self.shutDown)
        self.ui.getNextButton.clicked.connect(self.requestNext)
        self.ui.annButton.clicked.connect(self.annotateTest)
        self.ui.deferButton.clicked.connect(self.deferTest)
        self.ui.tagButton.clicked.connect(self.tagTest)
        self.ui.filterButton.clicked.connect(self.setFilter)
        self.ui.filterLE.returnPressed.connect(self.setFilter)
        self.ui.filterInvCB.stateChanged.connect(self.setFilter)
        self.ui.viewButton.clicked.connect(self.viewSpecificImage)

    def getMaxMark(self):
        """
        Get the max-mark for the question from the server.

        Returns
            True if max score retrieved successfully, False otherwise
        """
        try:
            self.maxMark = messenger.MgetMaxMark(self.question, self.version)
        except PlomRangeException as err:
            log.error(err)
            ErrorMessage(str(err)).exec_()
            self.shutDownError()
            return False
        except PlomSeriousException as err:
            self.throwSeriousError(err, rethrow=False)
            return False
        return True

    def setMarkStyleID(self):
        """
        Give IDs to the radio-buttons which select the marking style.

        Notes:
            Hides "mark total" style by default
            Mark style ID's are as follows
                1 = mark total = user clicks the total-mark, will likely be
                    removed in the future.
                2 = mark-up = mark starts at 0 and user increments it
                3 = mark-down = mark starts at max and user decrements it

        Returns:
            None

        """
        self.ui.markStyleGroup.setId(self.ui.markTotalRB, 1)
        self.ui.markTotalRB.hide()
        self.ui.markTotalRB.setEnabled(False)
        # continue with the other buttons
        self.ui.markStyleGroup.setId(self.ui.markUpRB, 2)
        self.ui.markStyleGroup.setId(self.ui.markDownRB, 3)

    def resizeEvent(self, event):
        """
        Resizes the image and surrounding table.

        Notes:
            Overrides QWidget.resizeEvent()
            a resize can be triggered before "getToWork" is called.

        Args:
            event (QEvent): the event to be resized.

        Returns:
            None

        """
        if hasattr(self, "testImg"):
            self.testImg.resetB.animateClick()
        if hasattr(self, "ui.tableView"):
            self.ui.tableView.resizeRowsToContents()
        super(MarkerClient, self).resizeEvent(event)

    def throwSeriousError(self, error, rethrow=True):
        """
        Logs an exception, pops up a dialog and shuts down.

        Args:
            error: the exception to be reraised
            rethrow: True if the only way to solve this error is to crash
                and shut down Plom. False if the exception can be handled in a
                way other than crashing, in which case it will initiate
                shutdown and not re-raise the exception (thus avoiding a crash)

        Returns:
            None

        """
        # automatically prints a stacktrace into the log!
        log.exception("A serious error has been detected")
        msg = 'A serious error has been thrown:\n"{}"'.format(error)
        if rethrow:
            msg += "\nProbably we will crash now..."
        else:
            msg += "\nShutting down Marker."
        ErrorMessage(msg).exec_()
        self.shutDownError()
        if rethrow:
            raise (error)

    def loadMarkedList(self):
        """
        Loads the list of previously marked papers into self.examModel

        Returns:
            None

        """
        # Ask server for list of previously marked papers
        markedList = messenger.MrequestDoneTasks(self.question, self.version)
        for x in markedList:
            # TODO: might not the "markedList" have some other statuses?
            self.examModel.addPaper(
                ExamQuestion(
                    x[0],
                    fnames=[],
                    stat="marked",
                    mrk=x[1],
                    mtime=x[2],
                    tags=x[3],
                    integrity_check=x[4],
                    image_md5s=[],
                )
            )

    def loadOriginalFiles(self, task):
        """
        Loads the original image files for a given task.

        Args:
            task (str): the task for the image files to be loaded from.
                Takes the form "q1234g9" = test 1234 question 9

        Returns:
            True/False

        Raises:
            SeriousError if requesting images throws a PlomSeriousException

        """
        if len(self.examModel.getOriginalFiles(task)) > 0:
            return True

        try:
            [image_metadata, anImage, plImage] = messenger.MrequestImages(
                task, self.examModel.getIntegrityCheck(task)
            )
        except (PlomTaskChangedError, PlomTaskDeletedError) as ex:
            # TODO: better action we can take here?
            ErrorMessage(
                '<p>The task "{}" has changed in some way by the manager; it '
                "may need to be remarked.</p>\n\n"
                '<p>Specifically, the server says: "{}"</p>\n\n'
                "<p>This is a rare situation; just in case, we'll now force a "
                "shutdown of your client.  Sorry.</p>".format(task, str(ex))
            ).exec_()
            # This would avoid seeing the crash dialog...
            # import sys
            # sys.exit(58)
            raise PlomSeriousException("Manager changed task") from ex
        except PlomSeriousException as e:
            self.throwSeriousError(e)
            return False

        paperDir = tempfile.mkdtemp(prefix=task + "_", dir=self.workingDirectory)
        log.debug("create paperDir {} for already-graded download".format(paperDir))

        # Image names = "<task>.<imagenumber>.<extension>"
        image_fnames = []
        for i, row in enumerate(image_metadata):
            tmp = os.path.join(self.workingDirectory, "{}.{}.image".format(task, i))
            image_fnames.append(tmp)
            im_bytes = messenger.MrequestOneImage(task, row[0], row[1])
            with open(tmp, "wb+") as fh:
                fh.write(im_bytes)
        md5List = [x[1] for x in image_metadata]
        self.examModel.setOriginalFilesAndMD5s(task, image_fnames, md5List)

        if anImage is None:
            return True

        self.examModel.setPaperDirByTask(task, paperDir)
        aname = os.path.join(paperDir, "G{}.png".format(task[1:]))
        pname = os.path.join(paperDir, "G{}.plom".format(task[1:]))
        with open(aname, "wb+") as fh:
            fh.write(anImage)
        with open(pname, "wb+") as fh:
            fh.write(plImage)
        self.examModel.setAnnotatedFile(task, aname, pname)
        return True

    def _updateImage(self, pr=0):
        """
        Updates the image if needed.

        Args:
            pr (str): prefix of image to be loaded

        Returns:
            None

        """
        if not self.loadOriginalFiles(self.prxM.getPrefix(pr)):
            return

        if self.prxM.getStatus(pr) in ("marked", "uploading...", "???"):
            self.testImg.updateImage(self.prxM.getAnnotatedFile(pr))
        else:
            self.testImg.updateImage(self.prxM.getOriginalFiles(pr))
        self.testImg.forceRedrawOrSomeBullshit()
        self.ui.tableView.setFocus()

    def updateProgress(self, val=None, maxm=None):
        """
        Updates the progress bar.

        Args:
            val (int): value for the progress bar
            maxm (int): maximum for the progress bar.

        Returns:
            None

        """

        if not val and not maxm:
            # ask server for progress update
            try:
                val, maxm = messenger.MprogressCount(self.question, self.version)
            except PlomSeriousException as err:
                log.exception("Serious error detected while updating progress")
                msg = 'A serious error happened while updating progress:\n"{}"'.format(
                    err
                )
                msg += "\nThis is not good: restart, report bug, etc."
                ErrorMessage(msg).exec_()
                return
        if maxm == 0:
            val, maxm = (0, 1)  # avoid (0, 0) indeterminate animation
            self.ui.mProgressBar.setFormat("No papers to mark")
            ErrorMessage("No papers to mark.").exec_()
        else:
            # Neither is quite right, instead, we cache on init
            self.ui.mProgressBar.setFormat(self.ui._cachedProgressFormatStr)
        self.ui.mProgressBar.setMaximum(maxm)
        self.ui.mProgressBar.setValue(val)

    def requestNext(self):
        """
        Ask server for unmarked paper, get file, add to list, update view.

        Retry a few times in case two clients are asking for same.
        Notes:
            Side effects: on success, updates the table of tasks
            TODO: return value on success?  Currently None.
            TODO: rationalize return values

        Returns:
            None

        Raises:
            error if getting task from messenger throws PlomSeriousException

        """
        attempts = 0
        while True:
            attempts += 1
            # little sanity check - shouldn't be needed.
            # TODO remove.
            if attempts > 5:
                return
            # ask server for task of next task
            try:
                task = messenger.MaskNextTask(self.question, self.version)
                if not task:
                    return False
            except PlomSeriousException as err:
                self.throwSeriousError(err)

            try:
                image_metadata, tags, integrity_check = messenger.MclaimThisTask(task)
                break
            except PlomTakenException as err:
                log.info("will keep trying as task already taken: {}".format(err))
                continue

        image_md5_list = [x[1] for x in image_metadata]
        # Image names = "<task>.<imagenumber>.<extension>"
        image_fnames = []
        for i, row in enumerate(image_metadata):
            # try-except? how does this fail?
            im_bytes = messenger.MrequestOneImage(task, row[0], row[1])
            tmp = os.path.join(self.workingDirectory, "{}.{}.image".format(task, i))
            image_fnames.append(tmp)
            with open(tmp, "wb+") as fh:
                fh.write(im_bytes)

        self.examModel.addPaper(
            ExamQuestion(
                task,
                image_fnames,
                image_md5s=image_md5_list,
                tags=tags,
                integrity_check=integrity_check,
            )
        )
        pr = self.prxM.rowFromTask(task)
        if pr is not None:
            # if newly-added row is visible, select it and redraw
            self.ui.tableView.selectRow(pr)
            self._updateImage(pr)
            # Clean up the table
            self.ui.tableView.resizeColumnsToContents()
            self.ui.tableView.resizeRowsToContents()

    def requestNextInBackgroundStart(self):
        """
        Requests the next TGV in the background.

        Returns:
            None

        """
        if self.backgroundDownloader:
            log.info(
                "Previous Downloader ({}) still here, waiting".format(
                    str(self.backgroundDownloader)
                )
            )
            # if prev downloader still going than wait.  might block the gui
            self.backgroundDownloader.wait()
        self.backgroundDownloader = BackgroundDownloader(self.question, self.version)
        self.backgroundDownloader.downloadSuccess.connect(
            self._requestNextInBackgroundFinished
        )
        self.backgroundDownloader.downloadNoneAvailable.connect(
            self.requestNextInBackgroundNoneAvailable
        )
        self.backgroundDownloader.downloadFail.connect(
            self.requestNextInBackgroundFailed
        )
        self.backgroundDownloader.start()

    def _requestNextInBackgroundFinished(
        self, task, fnames, image_md5s, tags, integrity_check
    ):
        """
        Adds paper to exam model once it's been requested.

        Args:
            task (str): the task name for the next test.
            fnames (list[str]): the file names for next test
            image_md5s (list[str]): the md5sums for the images
            tags (str): tags for the TGV.
            integrity_check (str): integrity check string for the underlying images (concat of their md5sums)

        Returns:
            None

        """
        self.examModel.addPaper(
            ExamQuestion(
                task,
                fnames,
                image_md5s=image_md5s,
                tags=tags,
                integrity_check=integrity_check,
            )
        )
        # Clean up the table
        self.ui.tableView.resizeColumnsToContents()
        self.ui.tableView.resizeRowsToContents()

    def requestNextInBackgroundNoneAvailable(self):
        """
        Empty.

        Notes:
            Keep this function here just in case we want to do something in the
            future.
        """
        pass

    def requestNextInBackgroundFailed(self, errmsg):
        """
        Sends an error message when requesting the next exam fails.

        Args:
            errmsg (str): Error message received.

        Returns:
            None

        """
        # TODO what should we do?  Is there a realistic way forward
        # or should we just die with an exception?
        ErrorMessage(
            "Unfortunately, there was an unexpected error downloading "
            "next paper.\n\n{}\n\n"
            "Please consider filing an issue?  I don't know if its "
            "safe to continue from here...".format(errmsg)
        ).exec_()

    def moveToNextUnmarkedTest(self, task=None):
        """
        Move the list to the next unmarked test, if possible.

        Args:
            task (str): the task number of the next unmarked test.

        Returns:
             True if move was successful, False if not, for any reason.
        """
        if self.backgroundDownloader:
            # Might need to wait for a background downloader.  Important to
            # processEvents() so we can receive the downloader-finished signal.
            # TODO: assumes the downloader tries to stay just one ahead.
            count = 0
            while self.backgroundDownloader.isRunning():
                time.sleep(0.05)
                self.Qapp.processEvents()
                count += 1
                if (count % 10) == 0:
                    log.info("waiting for downloader to fill table...")
                if count >= 100:
                    msg = SimpleMessage(
                        "Still waiting for downloader to get the next image.  "
                        "Do you want to wait a few more seconds?\n\n"
                        "(It is safe to choose 'no': the Annotator will simply close)"
                    )
                    if msg.exec_() == QMessageBox.No:
                        return False
                    count = 0
            self.Qapp.processEvents()

        # Move to the next unmarked test in the table.
        # Be careful not to get stuck in a loop if all marked
        prt = self.prxM.rowCount()
        if prt == 0:  # no tasks
            return False

        prstart = None
        if task:
            prstart = self.prxM.rowFromTask(task)
        if not prstart:
            # it might be hidden by filters
            prstart = 0  # put 'start' at row=0
        pr = prstart
        while self.prxM.getStatus(pr) in ["marked", "uploading...", "deferred", "???"]:
            pr = (pr + 1) % prt
            if pr == prstart:  # don't get stuck in a loop
                break
        if pr == prstart:
            return False  # have looped over all rows and not found anything.
        self.ui.tableView.selectRow(pr)
        return True

    def deferTest(self):
        """Mark test as "defer" - to be skipped until later."""
        if len(self.ui.tableView.selectedIndexes()):
            pr = self.ui.tableView.selectedIndexes()[0].row()
        else:
            return
        task = self.prxM.getPrefix(pr)
        if self.examModel.getStatusByTask(task) == "deferred":
            return
        if self.examModel.getStatusByTask(task) in ("marked", "uploading...", "???"):
            msg = ErrorMessage(
                "Cannot defer a marked test. We will change this in a future version."
            )
            msg.exec_()
            return
        self.examModel.deferPaper(task)

    def startTheAnnotator(self, initialData):
        """
        This fires up the annotation window for user annotation + marking.

        Args:
            initialData (list): containing things documented elsewhere
                in :func:`plom.client.annotator.Annotator.__init__`.

        Returns:
            None

        """
        mouseHand = 1 if self.ui.leftMouseCB.isChecked() else 0

        annotator = Annotator(
            self.ui.userLabel.text(),
            mouseHand,
            parentMarkerUI=self,
            initialData=initialData,
        )
        # run the annotator
        annotator.annotator_upload.connect(self.callbackAnnWantsUsToUpload)
        annotator.annotator_done_closing.connect(self.callbackAnnDoneClosing)
        annotator.annotator_done_reject.connect(self.callbackAnnDoneCancel)
        self.setEnabled(False)
        annotator.show()

        # TODO: the old one might still be closing when we get here, but dropping
        # the ref now won't hurt (I think).
        self._annotator = annotator

    def annotateTest(self):
        """Grab current test from table, do checks, start annotator."""
        if len(self.ui.tableView.selectedIndexes()):
            row = self.ui.tableView.selectedIndexes()[0].row()
        else:
            return

        task = self.prxM.getPrefix(row)
        inidata = self.getDataForAnnotator(task)
        # make sure getDataForAnnotator did not fail
        if inidata is None:
            return

        if self.allowBackgroundOps:
            if self.examModel.countReadyToMark() == 0:
                self.requestNextInBackgroundStart()

        self.startTheAnnotator(inidata)
        # we started the annotator, we'll get a signal back when its done

    def getDataForAnnotator(self, task):
        """
        Start annotator on a particular task.

        Args:
            task (str): the task id.  If original qXXXXgYY, then annotated
                version is GXXXXgYY (G=graded).
        Returns
            data (list): (as described by startTheAnnotator) if successful.
        """
        # Create annotated filename.
        assert task.startswith("q")
        Gtask = "G" + task[1:]
        paperdir = tempfile.mkdtemp(prefix=task[1:] + "_", dir=self.workingDirectory)
        log.debug("create paperdir {} for annotating".format(paperdir))
        aname = os.path.join(paperdir, Gtask + ".png")
        cname = os.path.join(paperdir, Gtask + ".json")
        pname = os.path.join(paperdir, Gtask + ".plom")

        remarkFlag = False

        if self.examModel.getStatusByTask(task) in ("marked", "uploading...", "???"):
            msg = SimpleMessage("Continue marking paper?")
            if not msg.exec_() == QMessageBox.Yes:
                return
            remarkFlag = True
            oldpaperdir = self.examModel.getPaperDirByTask(task)
            log.debug("oldpaperdir is " + oldpaperdir)
            assert oldpaperdir is not None
            oldaname = os.path.join(oldpaperdir, Gtask + ".png")
            oldpname = os.path.join(oldpaperdir, Gtask + ".plom")
            # TODO: comment json file not downloaded
            # https://gitlab.com/plom/plom/issues/415
            shutil.copyfile(oldaname, aname)
            shutil.copyfile(oldpname, pname)

        # Yes do this even for a regrade!  We will recreate the annotations
        # (using the plom file) on top of the original file.
        fnames = self.examModel.getOriginalFiles(task)
        if self.backgroundDownloader:
            count = 0
            # Notes: we could check using `while not os.path.exists(fname):`
            # Or we can wait on the downloader, which works when there is only
            # one download thread.  Better yet might be a dict/database that
            # we update on downloadFinished signal.
            while self.backgroundDownloader.isRunning():
                time.sleep(0.1)
                count += 1
                # if .remainder(count, 10) == 0: # this is only python3.7 and later. - see #509
                if (count % 10) == 0:
                    log.info("waiting for downloader: {}".format(fnames))
                if count >= 40:
                    msg = SimpleMessage(
                        "Still waiting for download.  Do you want to wait a bit longer?"
                    )
                    if msg.exec_() == QMessageBox.No:
                        return
                    count = 0

        # maybe the downloader failed for some (rare) reason
        for fn in fnames:
            if not os.path.exists(fn):
                log.warning(
                    "some kind of downloader fail? (unexpected, but probably harmless"
                )
                return

        # stash the previous state, not ideal because makes column wider
        prevState = self.examModel.getStatusByTask(task)
        self.examModel.setStatusByTask(task, "ann:" + prevState)

        if remarkFlag:
            with open(pname, "r") as fh:
                pdict = json.load(fh)
        else:
            pdict = None

        exam_name = self.exam_spec["name"]
        markStyle = self.ui.markStyleGroup.checkedId()
        tgv = task[1:]
        # get the integrity_check code and image_md5_list of the task
        integrity_check = self.examModel.getIntegrityCheck(task)
        # convert back to list[int]
        image_md5_list = json.loads(self.examModel.getImageMD5List(task))
        return (
            tgv,
            exam_name,
            paperdir,
            fnames,
            aname,
            self.maxMark,
            markStyle,
            pdict,
            integrity_check,
            image_md5_list,
        )

    # when Annotator done, we come back to one of these callbackAnnDone* fcns
    @pyqtSlot(str)
    def callbackAnnDoneCancel(self, task):
        """
        Called when anotator is done grading.

        Args:
            task (str): task name

        Returns:
            None

        """
        self.setEnabled(True)
        if task:
            prevState = self.examModel.getStatusByTask("q" + task).split(":")[-1]
            # TODO: could also erase the paperdir
            self.examModel.setStatusByTask("q" + task, prevState)
        # TODO: see below re "done grading".
        self._updateImage()

    @pyqtSlot(str)
    def callbackAnnDoneClosing(self, task):
        """
        Called when annotator is done grading and is closing.

        Args:
            task (str): the task ID of the current test.

        Returns:
            None

        """
        self.setEnabled(True)
        # update image view, if the row we just finished is selected
        prIndex = self.ui.tableView.selectedIndexes()
        if len(prIndex) == 0:
            return
        pr = prIndex[0].row()
        # TODO: when done grading, if ann stays open, then close, this doesn't happen
        if task:
            if self.prxM.getPrefix(pr) == "q" + task:
                self._updateImage(pr)

    @pyqtSlot(str, list)
    def callbackAnnWantsUsToUpload(self, task, stuff):
        """
        Called when annotator wants to upload.

        Args:
            task (str): the task ID of the current test.
            stuff (list): a list containing
                grade(int): grade given by marker.
                markingTime(int): total time spent marking.
                paperDir(dir): Working directory for the current task
                fnames(str): original file name (unannotated)
                aname(str): annotated file name
                plomFileName(str): the name of thee .plom file
                commentFileName(str): the name of the comment file.
                integrity_check(str): the integrity_check string of the task.
                image_md5_list(list[str]): the list of image md5sums

        Returns:
            None

        """
        (
            gr,
            markingTime,
            paperDir,
            fnames,
            aname,
            plomFileName,
            commentFileName,
            integrity_check,
            image_md5_list,
        ) = stuff

        if not (0 <= gr and gr <= self.maxMark):
            msg = ErrorMessage(
                "Mark of {} is outside allowed range. Rejecting. This should not happen. Please file a bug".format(
                    gr
                )
            )
            msg.exec_()
            # TODO: what do do here?  revert?
            return

        stat = self.examModel.getStatusByTask("q" + task)

        # Copy the mark, annotated filename and the markingtime into the table
        # TODO: sort this out whether task is "q00..." or "00..."?!
        self.examModel.markPaperByTask(
            "q" + task, gr, aname, plomFileName, markingTime, paperDir
        )
        # update the markingTime to be the total marking time
        totmtime = self.examModel.getMTimeByTask("q" + task)
        tags = self.examModel.getTagsByTask("q" + task)

        _data = (
            "q" + task,  # current task
            gr,  # grade
            (
                aname,
                plomFileName,
                commentFileName,
            ),  # annotated, plom, and comment filenames
            totmtime,  # total marking time (seconds)
            self.question,
            self.version,
            tags,
            integrity_check,
            image_md5_list,
        )
        if self.allowBackgroundOps:
            # the actual upload will happen in another thread
            self.backgroundUploader.enqueueNewUpload(*_data)
        else:
            upload(
                *_data,
                knownFailCallback=self.backgroundUploadFailedServerChanged,
                unknownFailCallback=self.backgroundUploadFailed,
                successCallback=self.backgroundUploadFinished,
            )

    def getMorePapers(self, oldtgvID):
        """
        Loads more tests.

        Args:
            oldtgvID(str): the Test-Group-Version ID for the previous test.

        Returns:
            initialData: as described by getDataForAnnotator

        """
        log.debug("Annotator wants more (w/o closing)")
        if not self.allowBackgroundOps:
            self.requestNext()
        if not self.moveToNextUnmarkedTest("q" + oldtgvID if oldtgvID else None):
            return False
        # TODO: copy paste of annotateTest()
        # probably don't need len check
        if len(self.ui.tableView.selectedIndexes()):
            row = self.ui.tableView.selectedIndexes()[0].row()
        else:
            # TODO: what to do here?
            return False
        tgvID = self.prxM.getPrefix(row)

        data = self.getDataForAnnotator(tgvID)
        # make sure getDataForAnnotator did not fail
        if data is None:
            return

        assert tgvID[1:] == data[0]
        pdict = data[-3]  # the plomdict is third-last object in data
        assert pdict is None, "Annotator should not pull a regrade"

        if self.allowBackgroundOps:
            # while annotator is firing up request next paper in background
            # after giving system a moment to do `annotator.exec_()`
            if self.examModel.countReadyToMark() == 0:
                self.requestNextInBackgroundStart()

        return data

    def PermuteAndGetSamePaper(self, task, imageList):
        """
        Allows user to reorganize pages of an exam.

        Args:
            task (str): the task ID of the current test.
            imageList (list[str]): list of image names to which are being
                rearranged.

        Returns:
            initialData (as described by getDataForAnnotator)

        """
        log.info("Rearranging image list for task {} = {}".format(task, imageList))
        # we know the list of image-refs and files. copy files into place
        # Image names = "<task>.<imagenumber>.<extension>"
        image_names = []
        image_md5s = []
        # TODO: This code was trying (badly) to overwrite the q0001 files...
        # TODO: something with tempfile instead
        rand6hex = secrets.token_hex(3)
        for i in range(len(imageList)):
            tmp = os.path.join(
                self.workingDirectory, "twist_{}_{}.{}.image".format(rand6hex, task, i)
            )
            shutil.copyfile(imageList[i][1], tmp)
            image_names.append(tmp)
            image_md5s.append(imageList[i][0])

        task = "q" + task
        self.examModel.setOriginalFilesAndMD5s(task, image_names, image_md5s)
        # set the status back to untouched so that any old plom files ignored
        self.examModel.setStatusByTask(task, "untouched")
        # finally relaunch the annotator
        return self.getDataForAnnotator(task)

    def backgroundUploadFinished(self, task, numDone, numtotal):
        """
        An upload has finished, do appropriate UI updates

        Args:
            task (str): the task ID of the current test.
            numDone (int): number of exams marked
            numTotal (int): total number of exams to mark.

        Returns:
            None

        """
        stat = self.examModel.getStatusByTask(task)
        # maybe it changed while we waited for the upload
        if stat == "uploading...":
            self.examModel.setStatusByTask(task, "marked")
        self.updateProgress(numDone, numtotal)

    def backgroundUploadFailedServerChanged(self, task, error_message):
        """An upload has failed because server changed something, safest to quit.

        Args:
            task (str): the task ID of the current test.
            error_message (str): a brief description of the error.

        Returns:
            None
        """
        self.examModel.setStatusByTask(task, "???")
        ErrorMessage(
            '<p>Background upload of "{}" has failed because the server '
            "changed something underneath us.</p>\n\n"
            '<p>Specifically, the server says: "{}"</p>\n\n'
            "<p>This is a rare situation; no data corruption has occured but "
            "your annotations have been discarded just in case.  You will be "
            "asked to redo the task later.</p>\n\n"
            "<p>For now you've been logged out and we'll now force a shutdown "
            "of your client.  Sorry.</p>".format(task, error_message)
        ).exec_()
        # This would avoid seeing the crash dialog...
        # import sys
        # sys.exit(57)
        raise PlomSeriousException(
            "Server changed under us: {}".format(error_message)
        ) from None

    def backgroundUploadFailed(self, task, errmsg):
        """An upload has failed, we don't know why, do something LOADLY.

        Args:
            task (str): the task ID of the current test.
            errmsg (str): the error message.

        Returns:
            None

        """
        self.examModel.setStatusByTask(task, "???")
        ErrorMessage(
            "Unfortunately, there was an unexpected error; the server did "
            "not accept our marked paper {}.\n\n{}\n\n"
            "If the problem persists consider filing an issue."
            "Please close this window and log in again.".format(task, errmsg)
        ).exec_()
        return

    def updateImg(self, newImg, oldImg):
        """
        Updates the displayed image when the selection has changed.

        Args:
            newImg (QItem): new image
            oldImg (QItem): old image

        Returns:
            None

        """
        idx = newImg.indexes()
        if len(idx) > 0:
            self._updateImage(idx[0].row())

    def shutDownError(self):
        """Shuts down self due to error."""
        if (
            getattr(self, "_annotator", None) is not None
        ):  # try to shut down annotator too.
            self._annotator.close()
        log.error("shutting down")
        self.my_shutdown_signal.emit(2, [])
        self.close()

    def shutDown(self):
        """Shuts down self."""
        log.debug("Marker shutdown from thread " + str(threading.get_ident()))
        if self.backgroundUploader:
            count = 42
            while self.backgroundUploader.isRunning():
                if self.backgroundUploader.isEmpty():
                    # don't try to quit until the queue is empty
                    self.backgroundUploader.quit()
                time.sleep(0.1)
                count += 1
                if count >= 50:
                    count = 0
                    msg = SimpleMessage(
                        "Still waiting for uploader to finish.  Do you want to wait a bit longer?"
                    )
                    if msg.exec_() == QMessageBox.No:
                        # politely ask one more time
                        self.backgroundUploader.quit()
                        time.sleep(0.1)
                        # then nuke it from orbit
                        if self.backgroundUploader.isRunning():
                            self.backgroundUploader.terminate()
                        break
            self.backgroundUploader.wait()

        # When shutting down, first alert server of any images that were
        # not marked - using 'DNF' (did not finish). Sever will put
        # those files back on the todo pile.
        self.DNF()
        # Then send a 'user closing' message - server will revoke
        # authentication token.
        try:
            messenger.closeUser()
        except PlomSeriousException as err:
            self.throwSeriousError(err)

        markStyle = self.ui.markStyleGroup.checkedId()
        mouseHand = 1 if self.ui.leftMouseCB.isChecked() else 0
        sidebarRight = self.ui.sidebarRightCB.isChecked()
        self.my_shutdown_signal.emit(2, [markStyle, mouseHand, sidebarRight])
        self.close()

    def DNF(self):
        """
        Marks files that are not finished as "did not finish."

        Notes:
            do this for everything, not just the proxy-model

        Returns:
            None

        Raises:
            PlomSeriousException if an error occurs in server.

        """
        for r in range(self.examModel.rowCount()):
            if self.examModel.data(self.examModel.index(r, 1)) != "marked":
                # Tell server the task fo any paper that is not marked.
                # server will put that back on the todo-pile.
                try:
                    messenger.MdidNotFinishTask(
                        self.examModel.data(self.examModel.index(r, 0))
                    )
                except PlomSeriousException as err:
                    self.throwSeriousError(err)

    def downloadWholePaper(self, testNumber):
        """

        Args:
            testNumber (int): the test number.

        Returns:
            (tuple) containing pageData and viewFiles


        """
        try:
            pageData, imagesAsBytes = messenger.MrequestWholePaper(
                testNumber, self.question
            )
        except PlomTakenException as err:
            log.exception("Taken exception when downloading whole paper")
            ErrorMessage("{}".format(err)).exec_()
            return ([], [])  # TODO: what to return?

        viewFiles = []
        for iab in imagesAsBytes:
            tfn = tempfile.NamedTemporaryFile(
                dir=self.workingDirectory, suffix=".image", delete=False
            ).name
            viewFiles.append(tfn)
            with open(tfn, "wb") as fh:
                fh.write(iab)

        return [pageData, viewFiles]

    def downloadWholePaperMetadata(self, testNumber):
        """Get metadata about all images used in a particular test paper.

        Args:
            testNumber (int): the test number.

        Returns:
            (tuple) containing pageData and viewFiles
        """
        pageData = messenger.MrequestWholePaperMetadata(testNumber, self.question)
        return pageData

    def downloadOneImage(self, task, image_id, md5):
        """Download one image from server by its database id."""
        return messenger.MrequestOneImage(task, image_id, md5)

    def doneWithWholePaperFiles(self, viewFiles):
        """ Unlinks files in viewFiles to os. """
        for f in viewFiles:
            if os.path.isfile(f):
                os.unlink(f)

    def cacheLatexComments(self):
        """Caches Latexed comments."""
        clist = commentLoadAll()
        # sort list in order of longest comment to shortest comment
        clist.sort(key=lambda C: -len(C["text"]))

        # Build a progress dialog to warn user
        pd = QProgressDialog("Caching latex comments", None, 0, 2 * len(clist), self)
        pd.setWindowModality(Qt.WindowModal)
        pd.setMinimumDuration(0)
        # Start caching.
        c = 0
        pd.setValue(c)
        n = int(self.question)
        exam_name = self.exam_spec["name"]

        for X in clist:
            if commentIsVisible(X, n, exam_name) and X["text"][:4].upper() == "TEX:":
                txt = X["text"][4:].strip()
                pd.setLabelText("Caching:\n{}".format(txt[:64]))
                # latex the red version
                self.latexAFragment(txt)
                c += 1
                pd.setValue(c)
                # and latex the preview
                txtp = "\\color{blue}" + txt  # make color blue for ghost rendering
                self.latexAFragment(txtp)
                c += 1
                pd.setValue(c)
            else:
                c += 2
                pd.setLabelText("Caching:\nno tex")
                pd.setValue(c)
        pd.close()

    def latexAFragment(self, txt):
        """
        Run LaTeX on a fragment of text and return the file name of a png.

        The files are cached for reuse if the same text is passed again.

        Args:
            txt (str): the text to be Latexed.

        Returns:
            (png): a file containing the Latexed text.

        """
        if txt in self.commentCache:
            # have already latex'd this comment
            return self.commentCache[txt]
        log.debug('requesting latex for "{}"'.format(txt))
        try:
            fragment = messenger.MlatexFragment(txt)
        except PlomLatexException:
            return None
        # a name for the fragment file
        fragFile = tempfile.NamedTemporaryFile(
            dir=self.workingDirectory, suffix=".png", delete=False
        ).name
        # save it
        with open(fragFile, "wb+") as fh:
            fh.write(fragment)
        # add it to the cache
        self.commentCache[txt] = fragFile
        return fragFile

    def tagTest(self):
        """ Adds a tag to the current Test."""
        if len(self.ui.tableView.selectedIndexes()):
            pr = self.ui.tableView.selectedIndexes()[0].row()
        else:
            return
        task = self.prxM.getPrefix(pr)
        tagSet = self.examModel.getAllTags()
        currentTag = self.examModel.getTagsByTask(task)

        atb = AddTagBox(self, currentTag, list(tagSet))
        if atb.exec_() == QDialog.Accepted:
            txt = atb.TE.toPlainText().strip()
            # truncate at 256 characters.  TODO: without warning?
            if len(txt) > 256:
                txt = txt[:256]

            self.examModel.setTagsByTask(task, txt)
            # resize view too
            self.ui.tableView.resizeRowsToContents()

            # send updated tag back to server.
            try:
                messenger.MsetTag(task, txt)
            except PlomTakenException as err:
                log.exception("exception when trying to set tag")
                ErrorMessage('Could not set tag:\n"{}"'.format(err)).exec_()
            except PlomSeriousException as err:
                self.throwSeriousError(err)
        return

    def setFilter(self):
        """ Sets a filter tag. """
        self.prxM.setFilterString(self.ui.filterLE.text().strip())
        # check to see if invert-filter is checked
        if self.ui.filterInvCB.checkState() == Qt.Checked:
            self.prxM.filterTags(invert=True)
        else:
            self.prxM.filterTags()

    def viewSpecificImage(self):
        """ shows the image.  """
        if self.canViewAll:
            tgs = SelectTestQuestion(self.exam_spec, self.question)
            if tgs.exec_() == QDialog.Accepted:
                tn = tgs.tsb.value()
                gn = tgs.gsb.value()
            else:
                return
        else:
            tgs = SelectTestQuestion(self.exam_spec)
            if tgs.exec_() == QDialog.Accepted:
                tn = tgs.tsb.value()
                gn = self.question
            else:
                return
        task = "q{}g{}".format(str(tn).zfill(4), int(self.question))
        try:
            imageList = messenger.MrequestOriginalImages(task)
        except PlomNoMoreException as err:
            msg = ErrorMessage("No image corresponding to task {}".format(task))
            msg.exec_()
            return
        ifilenames = []
        for img in imageList:
            ifile = tempfile.NamedTemporaryFile(
                dir=self.workingDirectory, suffix=".image", delete=False
            )
            ifile.write(img)
            ifilenames.append(ifile.name)
        tvw = GroupView(ifilenames)
        tvw.setWindowTitle(
            "Original ungraded image for question {} of test {}".format(gn, tn)
        )
        tvw.exec_()



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2020 Andrew Rechnitzer
# Copyright (C) 2020 Colin B. Macdonald

from datetime import datetime
import logging

import peewee as pw

from plom.rules import censorStudentNumber as censorID
from plom.rules import censorStudentName as censorName
from plom.db.tables import *


log = logging.getLogger("DB")


# ------------------
# Identifier stuff
# The ID-able tasks have group_type ="i", group.scanned=True,
# The todo id-tasks are IDGroup.status="todo"
# the done id-tasks have IDGroup.status="done"


def IDcountAll(self):
    """Count all tests in which ID pages are scanned."""
    try:
        return (
            Group.select()
            .where(
                Group.group_type == "i",
                Group.scanned == True,
            )
            .count()
        )
    except Group.DoesNotExist:
        return 0


def IDcountIdentified(self):
    """Count all tests in which ID pages are scanned and student has been identified."""
    try:
        return (
            IDGroup.select()
            .join(Group)
            .where(
                Group.scanned == True,
                IDGroup.identified == True,
            )
            .count()
        )
    except IDGroup.DoesNotExist:
        return 0


def IDgetNextTask(self):
    """Find unid'd test and send test_number to client"""
    with plomdb.atomic():
        try:  # grab the IDData reference provided not IDd but has been scanned
            iref = (
                IDGroup.select()
                .join(Group)
                .where(
                    IDGroup.status == "todo",
                    Group.scanned == True,
                )
                .get()
            )
            # note - test need not be all scanned, just the ID pages.
        except IDGroup.DoesNotExist:
            log.info("Nothing left on ID to-do pile")
            return None

        log.debug("Next ID task = {}".format(iref.test.test_number))
        return iref.test.test_number


def IDgiveTaskToClient(self, user_name, test_number):
    """Assign test #test_number as a task to the given user. Provided that task has not already been taken by another user, we return [True, image-list]."""
    uref = User.get(name=user_name)
    # since user authenticated, this will always return legit ref.
    with plomdb.atomic():
        # get that test
        tref = Test.get_or_none(Test.test_number == test_number)
        if (
            tref is None
        ):  # should not happen - user should not be asking for nonexistant tests
            log.info("ID task - That test number {} not known".format(test_number))
            return [False]
        # grab the ID group of that test
        iref = tref.idgroups[0]
        # verify the id-group has been scanned - it should be if we got here.
        if iref.group.scanned == False:
            return [False]
        if not (iref.user is None or iref.user == uref):
            # has been claimed by someone else.
            return [False]
        # update status, owner of task, time
        iref.status = "out"
        iref.user = uref
        iref.time = datetime.now()
        iref.save()
        # update user activity
        uref.last_action = "Took ID task {}".format(test_number)
        uref.last_activity = datetime.now()
        uref.save()
        # return [true, page1, page2, etc]
        rval = [True]
        for p in iref.idpages.order_by(IDPage.order):
            rval.append(p.image.file_name)

        log.debug("Giving ID task {} to user {}".format(test_number, user_name))
        return rval


def IDgetDoneTasks(self, user_name):
    """When a id-client logs on they request a list of papers they have already IDd.
    Send back the list."""
    uref = User.get(name=user_name)
    # since user authenticated, this will always return legit ref.

    query = IDGroup.select().where(IDGroup.user == uref, IDGroup.status == "done")
    idList = []
    for iref in query:
        idList.append([iref.test.test_number, iref.student_id, iref.student_name])
    log.debug("Sending completed ID tasks to user {}".format(user_name))
    return idList


def IDgetImage(self, user_name, test_number):
    """Return ID page images (+ Lpages) of test #test_number to user."""
    uref = User.get(name=user_name)
    # since user authenticated, this will always return legit ref.

    tref = Test.get_or_none(Test.test_number == test_number)
    if tref is None:
        return [False, "NoTest"]
    # grab the IDData
    iref = tref.idgroups[0]
    # check corresponding group is scanned
    if iref.group.scanned is False:
        return [False, "NoScan"]
    # quick sanity check to make sure task given to user, (or if manager making request)
    if iref.user != uref and user_name != "manager":
        return [False, "NotOwner"]
    rval = [True]
    for p in iref.idpages.order_by(IDPage.order):
        rval.append(p.image.file_name)
    log.debug("Sending IDpages of test {} to user {}".format(test_number, user_name))
    return rval


def IDgetImageByNumber(self, image_number):
    """
    For every test, find the imageNumber'th page in the ID Pages and return the corresponding image filename. So gives returns a dictionary of testNumber -> filename.
    """
    rval = {}
    query = IDGroup.select()
    for iref in query:
        # for each iref, check that it is scanned and then grab page.
        gref = iref.group
        if not gref.scanned:
            continue
        # make a list of all the pages in the IDgroup
        pages = []
        for p in iref.idpages:
            pages.append(p.image.file_name)
        # grab the relevant page if it is in the list
        if len(pages) > image_number:
            rval[iref.test.test_number] = pages[image_number]
        # otherwise we don't add that test to the dictionary.
    return rval


def IDdidNotFinish(self, user_name, test_number):
    """When user logs off, any images they have still out should be put
    back on todo pile
    """
    uref = User.get(name=user_name)
    # since user authenticated, this will always return legit ref.

    # Log user returning given tgv.
    with plomdb.atomic():
        tref = Test.get_or_none(Test.test_number == test_number)
        if tref is None:
            log.info("That test number {} not known".format(test_number))
            return False

        iref = tref.idgroups[0]
        # sanity check that user has task out.
        if iref.user != uref or iref.status != "out":
            return False
        # update status, id-time. If out, then student name/number are not set.
        iref.status = "todo"
        iref.user = None
        iref.time = datetime.now()
        iref.identified = False
        iref.save()
        tref.identified = False
        tref.save()
        log.info("User {} did not ID task {}".format(user_name, test_number))


def ID_id_paper(self, paper_num, user_name, sid, sname, checks=True):
    """Associate student name and id with a paper in the database.

    See also :func:`plom.db.db_create.id_paper` which is just this with
    `checks=False`.

    Args:
        paper_num (int)
        user_name (str): User who did the IDing.
        sid (str, None): student ID.  `None` if the ID page was blank:
            typically `sname` will then contain some short explanation.
        sname (str): student name.
        checks (bool): by default (True), the paper must be scanned
            and the `username` must match the current owner of the
            paper (typically because the paper was assigned to them).
            You can pass False if its ID the paper without being
            owner (e.g., during automated IDing of prenamed papers.)

    Returns:
        tuple: `(True, None, None)` if succesful, `(False, 409, msg)`
            means `sid` is in use elsewhere, a serious problem for
            the caller to deal with.  `(False, int, msg)` covers all
            other errors.  `msg` gives details about errors.  Some
            of these should not occur, and indicate possible bugs.
            `int` gives a hint of suggested HTTP status code,
            currently it can be 404, 403, or 409.

    TODO: perhaps several sorts of exceptions would be better.
    """
    uref = User.get(name=user_name)
    # since user authenticated, this will always return legit ref.

    logbase = 'User "{}" tried to ID paper {}'.format(user_name, paper_num)
    with plomdb.atomic():
        tref = Test.get_or_none(Test.test_number == paper_num)
        if tref is None:
            msg = "denied b/c paper not found"
            log.error("{}: {}".format(logbase, msg))
            return False, 404, msg
        iref = tref.idgroups[0]
        if checks and iref.group.scanned == False:
            msg = "denied b/c its not scanned yet"
            log.error("{}: {}".format(logbase, msg))
            return False, 404, msg
        if checks and iref.user != uref:
            msg = 'denied b/c it belongs to user "{}"'.format(iref.user)
            log.error("{}: {}".format(logbase, msg))
            return False, 403, msg
        iref.user = uref
        iref.status = "done"
        iref.student_id = sid
        iref.student_name = sname
        iref.identified = True
        iref.time = datetime.now()
        try:
            iref.save()
        except pw.IntegrityError:
            msg = "student id {} already entered elsewhere".format(censorID(sid))
            log.error("{} but {}".format(logbase, msg))
            return False, 409, msg
        tref.identified = True
        tref.save()
        # update user activity
        uref.last_action = "Returned ID task {}".format(paper_num)
        uref.last_activity = datetime.now()
        uref.save()
        if sid:
            log.info(
                'Paper {} ID\'d by "{}" as "{}" "{}"'.format(
                    paper_num, user_name, censorID(sid), censorName(sname)
                )
            )
        else:
            log.info(
                'Paper {} ID\'d by "{}" as "{}" "{}"'.format(
                    paper_num, user_name, sid, sname
                )
            )
    return True, None, None


def IDgetImageFromATest(self):
    """Returns ID images from the first unid'd test."""
    query = (  # look for scanned ID groups which are not IDd yet.
        IDGroup.select()
        .join(Group)
        .where(
            Group.group_type == "i",
            Group.scanned == True,
            IDGroup.identified == False,
        )
        .limit(1)  # we only need 1.
    )
    if query.count() == 0:
        log.info("No unIDd IDPages to sennd to manager")
        return [False]
    log.info("Sending first unIDd IDPages to manager")

    iref = query[0]
    rval = [True]
    for p in iref.idpages.order_by(IDPage.order):
        rval.append(p.image.file_name)

    return rval


def IDreviewID(self, test_number):
    """Replace the owner of the ID task for test test_number, with the reviewer."""
    # shift ownership to "reviewer"
    revref = User.get(name="reviewer")  # should always be there

    tref = Test.get_or_none(Test.test_number == test_number)
    if tref is None:
        return [False]
    iref = IDGroup.get_or_none(
        IDGroup.test == tref,
        IDGroup.identified == True,
    )
    if iref is None:
        return [False]
    with plomdb.atomic():
        iref.user = revref
        iref.time = datetime.now()
        iref.save()
    log.info("ID task {} set for review".format(test_number))
    return [True]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Andrew Rechnitzer"
__copyright__ = "Copyright (C) 2018-2020 Andrew Rechnitzer"
__credits__ = ["Andrew Rechnitzer", "Colin Macdonald", "Elvis Cai", "Matt Coles"]
__license__ = "AGPL-3.0-or-later"
# SPDX-License-Identifier: AGPL-3.0-or-later

import os
import sys
from time import localtime
from PyQt5.QtGui import QBrush, QPixmap


class BackGrid(QBrush):
    def __init__(self, username=None):
        # init the qgraphicsview
        super(BackGrid, self).__init__()
        # set the area outside the groupimage to be tiled grid png
        try:
            base_path = sys._MEIPASS
        except Exception:
            base_path = os.path.dirname(__file__)
        if (localtime().tm_mon == 4 and localtime().tm_mday == 1) or (
            username
            and any(x in username.lower() for x in ["omer", "remo", "legna", "angel"])
        ):
            self.setTexture(QPixmap(os.path.join(base_path, "backGrid2.png")))
        # else:
        #    self.setTexture(QPixmap(os.path.join(base_path, "backGrid1.svg")))



from plom.db.tables import *
from datetime import datetime, timedelta

import logging

log = logging.getLogger("DB")

# ------------------
# Reporting functions


def RgetScannedTests(self):
    """Get a dict of all scanned tests indexed by test_number.
    Each test lists pairs [page-code, page-version].
    page-code is t{page}, h{question}{order}, or l{order}.
    """
    scan_dict = {}
    for tref in Test.select().where(Test.scanned == True):
        pScanned = []
        # first append test-pages
        for p in tref.tpages:
            if p.scanned == True:
                pScanned.append(["t.{}".format(p.page_number), p.version])
        # then append hw-pages in question-order
        for qref in tref.qgroups:
            gref = qref.group
            for p in gref.hwpages:
                pScanned.append(["h.{}.{}".format(qref.question, p.order), p.version])
        # then append extra-pages in question-order
        for qref in tref.qgroups:
            gref = qref.group
            for p in gref.expages:
                pScanned.append(["e.{}.{}".format(qref.question, p.order), p.version])
        # then append loose-pages in order
        for p in tref.lpages:
            pScanned.append(["l.{}".format(p.order), 0])  # we don't know the version
        scan_dict[tref.test_number] = pScanned
    log.debug("Sending list of scanned tests")
    return scan_dict


def RgetIncompleteTests(self):
    """Get dict of incomplete tests - ie some test pages scanned but not all.
    Indexed by test_number
    Each test lists triples [page-code, page-version, scanned_or_not].
    page-code is t{page}, h{question}{order}, or l{order}.
    Note - if no tpages scanned, then it will not return tpages.
    Similalry, if no hwpages/expages scanned, then it will not return hwpages/expages.
    """
    incomp_dict = {}
    for tref in Test.select().where(Test.scanned == False, Test.used == True):
        page_state = []
        # if no tpages scanned then don't display
        if TPage.select().where(TPage.test == tref, TPage.scanned == True).count() > 0:
            for p in tref.tpages:
                page_state.append(["t.{}".format(p.page_number), p.version, p.scanned])

        # if no HW pages at all - then don't display.
        if tref.hwpages.count() > 0:
            # then append hw-pages in question-order
            for qref in tref.qgroups.order_by(QGroup.question):
                # if no HW pages scanned then display a hwpage 1 as unscanned.
                if qref.group.hwpages.count() == 0:
                    page_state.append(
                        ["h.{}.{}".format(qref.question, 1), qref.version, False]
                    )
                else:
                    for p in qref.group.hwpages:  # hw pages are always scanned
                        page_state.append(
                            ["h.{}.{}".format(qref.question, p.order), p.version, True]
                        )
        # if no ex pages at all - then don't display.
        if tref.expages.count() > 0:
            # then append ex-pages in question-order
            for qref in tref.qgroups.order_by(QGroup.question):
                for p in qref.group.expages:  # ex pages are always scanned
                    page_state.append(
                        ["ex.{}.{}".format(qref.question, p.order), p.version, True]
                    )
        # then append l-pages in order
        for p in tref.lpages:
            page_state.append(["l.{}".format(p.order), 0, True])
            # we don't know the version
        incomp_dict[tref.test_number] = page_state
    log.debug("Sending list of incomplete tests")
    return incomp_dict


def RgetCompleteHW(self):
    """Get a list of [test_number, sid] that have complete hw-uploads - ie all questions present."""
    hw_complete = []
    # look at all the scanned tests - they will either be hwpages or tpages
    for tref in Test.select().where(Test.scanned == True):
        # note - skip those with scanned TPages present.
        if TPage.get_or_none(test=tref, scanned=True) is None:
            hw_complete.append([tref.test_number, tref.idgroups[0].student_id])
    return hw_complete


def RgetMissingHWQ(self):
    """Get dict of tests with missing HW Pages - ie some test pages scanned but not all.
    Indexed by test_number
    Each test gives [scanned-tpages-present boolean, sid, missing-question-numbers].
    """
    incomp_dict = {}
    # look at tests that are not completely scanned
    for tref in Test.select().where(
        Test.scanned == False, Test.used == True, Test.identified == True
    ):
        # check if that test has any scanned tpages
        if TPage.get_or_none(test=tref, scanned=True) is None:
            question_list = [False, tref.idgroups[0].student_id]
        else:
            question_list = [True, tref.idgroups[0].student_id]
        for qref in tref.qgroups.order_by(QGroup.question):
            # if no HW pages scanned then display a hwpage 1 as unscanned.
            if qref.group.hwpages.count() == 0:
                question_list.append(qref.question)
        if len(question_list) > 1:
            incomp_dict[tref.test_number] = question_list
    log.debug("Sending list of missing hw questions")
    return incomp_dict


def RgetUnusedTests(self):
    """Return list of tests (by testnumber) that have not been used - ie no test-pages scanned, no hw pages scanned, no loose pages scanned."""
    unused_list = []
    for tref in Test.select().where(Test.used == False):
        unused_list.append(tref.test_number)
    log.debug("Sending list of unused tests")
    return unused_list


def RgetIdentified(self):
    """
    Return dict of identified tests - ie ones for which student ID/name are known.
    Indexed by test-number, lists pairs (student_id/student_name).
    Note that this includes papers which are not completely scanned.
    """
    idd_dict = {}
    for iref in IDGroup.select().where(IDGroup.identified == True):
        idd_dict[iref.test.test_number] = (iref.student_id, iref.student_name)
    log.debug("Sending list of identified tests")
    return idd_dict


def RgetProgress(self, q, v):
    """For the given question/version return a simple progress summary = a dict with keys
    [numberScanned, numberMarked, numberRecent, avgMark, avgTimetaken] and their values
    numberRecent = number done in the last hour.
    """
    # set up a time-delta of 1 hour for calc of number done recently.
    one_hour = timedelta(hours=1)

    NScanned = 0  # number scanned
    NMarked = 0  # number marked
    NRecent = 0  # number marked in the last hour
    SMark = 0  # sum mark - for computing average
    SMTime = 0  # sum marking time - for computing average

    for qref in (
        QGroup.select()
        .join(Group)
        .where(
            QGroup.question == q,
            QGroup.version == v,
            Group.scanned == True,
        )
    ):
        NScanned += 1
        if qref.marked == True:
            NMarked += 1
            SMark += qref.annotations[-1].mark
            SMTime += qref.annotations[-1].marking_time
            if datetime.now() - qref.annotations[-1].time < one_hour:
                NRecent += 1

    log.debug("Sending progress summary for Q{}v{}".format(q, v))
    if NMarked == 0:  # in case nothing done.
        return {
            "NScanned": NScanned,
            "NMarked": NMarked,
            "NRecent": NRecent,
            "avgMark": None,
            "avgMTime": None,
        }
    else:
        return {
            "NScanned": NScanned,
            "NMarked": NMarked,
            "NRecent": NRecent,
            "avgMark": SMark / NMarked,
            "avgMTime": SMTime / NMarked,
        }


def RgetMarkHistogram(self, q, v):
    """Return a dict of dicts containing histogram of marks for the given q/v as hist[user][question][mark]=count."""
    histogram = {}
    for qref in (
        QGroup.select()
        .join(Group)
        .where(
            QGroup.question == q,
            QGroup.version == v,
            QGroup.marked == True,
            Group.scanned == True,
        )
    ):
        # make sure user.name in histogram
        if qref.user.name not in histogram:
            histogram[qref.user.name] = {}
        # make sure the mark is in the dict for that user
        if qref.annotations[-1].mark not in histogram[qref.user.name]:
            histogram[qref.user.name][qref.annotations[-1].mark] = 0
        # add to the count.
        histogram[qref.user.name][qref.annotations[-1].mark] += 1
    log.debug("Sending mark histogram for Q{}v{}".format(q, v))
    return histogram


def RgetMarked(self, q, v):
    """Return a list of all marked tasks with that q/v."""
    marked_list = []
    for qref in (
        QuestionData.select()
        .join(Group)
        .where(
            QuestionData.questionNumber == q,
            QuestionData.version == v,
            QuestionData.marked == True,
            Group.scanned == True,  # this might be redundant.
        )
    ):
        marked_list.append(qref.group.gid)
    log.debug("Sending list of marked tasks for Q{}V{}".format(q, v))
    return marked_list


def RgetQuestionUserProgress(self, q, v):
    """For the given q/v return the number of questions marked by each user (who marked something in this q/v - so no zeros).
    Return a list of the form [ number_scanned, [user, nmarked], [user, nmarked], etc]
    """
    user_counts = {}
    number_scanned = 0
    for qref in (
        QGroup.select()
        .join(Group)
        .where(
            QGroup.question == q,
            QGroup.version == v,
            Group.scanned == True,
        )
    ):
        number_scanned += 1
        if qref.marked == True:
            if qref.user.name not in user_counts:
                user_counts[qref.user.name] = 0
            user_counts[qref.user.name] += 1
    # build return list
    progress = [number_scanned]
    for user in user_counts:
        progress.append([user, user_counts[user]])
    log.debug("Sending question/user progress for Q{}v{}".format(q, v))
    return progress


def RgetCompletionStatus(self):
    """Return a dict of every scanned test (ie all test pages present). Each dict entry is of the form dict[test_number] = [identified_or_not, number_of_questions_marked]"""
    progress = {}
    for tref in Test.select().where(Test.scanned == True):
        number_marked = (
            QGroup.select().where(QGroup.test == tref, QGroup.marked == True).count()
        )
        progress[tref.test_number] = [tref.identified, number_marked]
    log.debug("Sending list of completed tests")
    return progress


def RgetOutToDo(self):
    """Return a list of tasks that are currently out with clients. These have status "todo".
    For each task we return a triple of [code, user, time]
    code = id-t{testnumber} or mrk-t{testnumber}-q{question}-v{version}
    note that the datetime object is not jsonable, so we format it using strftime.
    """
    # note - have to format the time as string since not jsonable.
    # x.time.strftime("%y:%m:%d-%H:%M:%S"),

    out_tasks = []
    for iref in IDGroup.select().where(IDGroup.status == "out"):
        out_tasks.append(
            [
                "id-t{}".format(iref.test.test_number),
                iref.user.name,
                iref.time.strftime("%y:%m:%d-%H:%M:%S"),
            ]
        )
    for qref in QGroup.select().where(QGroup.status == "out"):
        out_tasks.append(
            [
                "mrk-t{}-q{}-v{}".format(
                    qref.test.test_number, qref.question, qref.version
                ),
                qref.user.name,
                qref.time.strftime("%y:%m:%d-%H:%M:%S"),
            ]
        )
    log.debug("Sending list of tasks that are still out")
    return out_tasks


def RgetStatus(self, test_number):
    """For the given test_number return detailed status information.
    Return a dict containing keys and values
    * number = test_number
    * identified = id'd or not (boolean)
    * marked = marked or not (boolean)
    Then if id'd we also add keys/values
    * sid = student id
    * sname = student name
    * iwho = who did the id-ing.
    For each question then add a sub-dict with key = that question number, and key/values
    * marked = marked or not
    * version = the version of that question
    if marked also add
    * mark = the score
    * who = who did the marking.
    """
    tref = Test.get_or_none(Test.test_number == test_number)
    if tref is None:
        return [False]
    state = {
        "number": tref.test_number,
        "identified": tref.identified,
        "marked": tref.marked,
    }
    if tref.identified:
        iref = tref.idgroups[0]
        state["sid"] = iref.student_id
        state["sname"] = iref.student_name
        state["iwho"] = iref.user.name
    for qref in tref.qgroups:
        if qref.marked:
            state[qref.question] = {
                "marked": True,
                "version": qref.version,
                "mark": qref.annotations[-1].mark,
                "who": qref.annotations[-1].user.name,
            }
        else:
            state[qref.question] = {
                "marked": False,
                "version": qref.version,
            }

    log.debug("Sending status of test {}".format(test_number))
    return [True, state]


def RgetSpreadsheet(self):
    """Return a dict that contains all the information needed to build the spreadsheet."""
    # build a spreadsheet dict indexed by test_number
    # each value that dict is a dict which contains the info about that test
    sheet = {}
    # look for all tests that are completely scanned.
    for tref in Test.select().where(Test.scanned == True):
        # a dict for the current test.
        this_test = {
            "identified": tref.identified,  # id'd or not
            "marked": tref.marked,  # completely marked or not
            "sid": "",  # blank entry for student id - replaced if id'd
            "sname": "",  # blank entry for student name - replaced if id'd
        }
        # if identified update sid, sname.
        iref = tref.idgroups[0]
        if tref.identified:  # set the sid and sname.
            this_test["sid"] = iref.student_id
            this_test["sname"] = iref.student_name
        # check each question (in order)
        for qref in tref.qgroups.order_by(QGroup.question):
            # store the version and mark
            this_test["q{}v".format(qref.question)] = qref.version
            this_test["q{}m".format(qref.question)] = ""  # blank unless marked
            if qref.marked:  # if marked, updated.
                this_test["q{}m".format(qref.question)] = qref.annotations[-1].mark
        # insert the data for this_test into the spreadsheet dict.
        sheet[tref.test_number] = this_test
    log.debug("Sending spreadsheet data.")
    return sheet


def RgetOriginalFiles(self, test_number):
    """Return list of the filenames for the original (unannotated) page images for the given test."""
    page_files = []
    tref = Test.get_or_none(test_number=test_number)
    if tref is None:
        return []
    # append tpages, hwpages and then lpages.
    for pref in tref.tpages.order_by(TPage.page_number):
        page_files.append(pref.image.file_name)
    for qref in tref.qgroups.order_by(QGroup.question):
        for pref in qref.group.hwpages:
            page_files.append(pref.image.file_name)
    for lref in tref.lpages.order_by(LPage.order):
        page_files.append(pref.image.file_name)

    log.debug("Sending original images of test {}".format(test_number))
    return page_files


def RgetCoverPageInfo(self, test_number):
    """For the given test, return information to build the coverpage for the test.
    We return a list of the form
    [[student_id, student_name], [question, version, mark]-for each question]
    """
    # todo - put in sanity / safety checks

    tref = Test.get_or_none(test_number=test_number)
    if tref is None:
        return []
    # [ID, Name]
    iref = tref.idgroups[0]
    coverpage = [[iref.student_id, iref.student_name]]
    # then [q, v, mark]
    for qref in tref.qgroups.order_by(QGroup.question):
        coverpage.append([qref.question, qref.version, qref.annotations[-1].mark])
    log.debug("Sending coverpage info of test {}".format(test_number))
    return coverpage


def RgetAnnotatedFiles(self, test_number):
    """For the given test return a list of the image file names for the idgroup, dnmgroup and the (marked) questions."""
    # todo - put in sanity / safety checks - making sure questions are marked.

    image_list = []
    tref = Test.get_or_none(test_number=test_number)
    if tref is None:
        return []
    # append ID-pages, then DNM-pages, then QuestionGroups
    idref = IDGroup.get_or_none(test=tref)
    for p in idref.idpages.order_by(IDPage.order):
        image_list.append(p.image.file_name)
    # append DNM pages
    dnmref = DNMGroup.get_or_none(test=tref)
    for p in dnmref.dnmpages.order_by(DNMPage.order):
        image_list.append(p.image.file_name)
    # append last annotation from each qgroup
    for g in tref.qgroups.order_by(QGroup.question):
        image_list.append(g.annotations[-1].aimage.file_name)
    log.debug("Sending annotated images for test {}".format(test_number))
    return image_list


def RgetMarkReview(self, filterQ, filterV, filterU):
    """Return a list of all marked qgroups satisfying the filter conditions.
    Filter on question-number, version, and user-name.
    For each matching qgroup we return a tuple of
    [testnumber, question, version, mark of latest annotation, username, marking_time, time finished.]
    """
    query = QGroup.select().join(User).where(QGroup.marked == True)
    if filterQ != "*":
        query = query.where(QGroup.question == filterQ)
    if filterV != "*":
        query = query.where(QGroup.version == filterV)
    if filterU != "*":
        query = query.where(User.name == filterU)
    filtered = []
    for qref in query:
        filtered.append(
            [
                qref.test.test_number,
                qref.question,
                qref.version,
                qref.annotations[-1].mark,
                qref.user.name,
                qref.annotations[-1].marking_time,
                # CANNOT JSON DATETIMEFIELD.
                qref.annotations[-1].time.strftime("%y:%m:%d-%H:%M:%S"),
            ]
        )
    log.debug(
        "Sending filtered mark-review data. filters (Q,V,U)={}.{}.{}".format(
            filterQ, filterV, filterU
        )
    )
    return filtered


def RgetAnnotatedImage(self, test_number, question, version):
    """Return the filename of the annotated image for the given test/question/version."""
    tref = Test.get_or_none(test_number=test_number)
    if tref is None:  # sanity check
        return [False]
    qref = QGroup.get_or_none(
        QGroup.test == tref,
        QGroup.question == question,
        QGroup.version == version,
        QGroup.marked == True,
    )
    if qref is None:  # another sanity check.
        return [False]
    log.debug(
        "Sending annotated image of tqv {}.{}.{}".format(test_number, question, version)
    )
    return [True, qref.annotations[-1].aimage.file_name]


def RgetIDReview(self):
    """Return information about every identified paper.
    For each paper return a tuple of [test_number, who did the iding, the time, the student ID, and the student name]
    """
    id_paper_list = []
    query = IDGroup.select().where(IDGroup.identified == True)
    for iref in query:
        id_paper_list.append(
            [
                iref.test.test_number,
                iref.user.name,
                iref.time.strftime("%y:%m:%d-%H:%M:%S"),
                iref.student_id,
                iref.student_name,
            ]
        )
    log.debug("Sending ID review data")
    return id_paper_list


def RgetUserFullProgress(self, user_name):
    """Return the number of completed tasks of teach type for the given user.
    Return [ number_id'd, number_marked]
    number_marked = number marked for all questions.
    """
    uref = User.get_or_none(name=user_name)
    if uref is None:
        return []
    # return [#IDd, #marked]
    log.debug("Sending user {} progress data".format(user_name))
    return [
        IDGroup.select()
        .where(IDGroup.user == uref, IDGroup.identified == True)
        .count(),
        QGroup.select().where(QGroup.user == uref, QGroup.marked == True).count(),
    ]



#!/usr/bin/env python3
# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2020 Colin B. Macdonald
# Copyright (C) 2020 Dryden Wiebe

"""Read in a Canvas-exported spreadsheet and prepare marks for upload.

This will create a new csv file that you can manually upload/import
back to Canvas.
"""

__copyright__ = "Copyright (C) 2019-2020 Colin B. Macdonald and others"
__credits__ = ["The Plom Project Developers"]
__license__ = "AGPL-3.0-or-later"

import argparse

from plom.finish import make_canvas_gradefile

canvas_fromfile = "canvas_latest_export.csv"
canvas_grades_tofile = "canvas_grades_for_import.csv"
# TODO: check if former exists and latter does not, and give some
# basic instructions

# TODO: should get this from project?!
Default_canvas_test_name = "Midterm ("  # almost certainly wrong


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-p",
        type=str,
        metavar="PART_COLUMN_NAME",
        help='Partial Canvas column name of test, as detailed as possible (defaults to "{}", see docs)'.format(
            Default_canvas_test_name
        ),
    )

    args = parser.parse_args()

    if not args.p:
        canvas_test_name = Default_canvas_test_name
    else:
        canvas_test_name = args.p

    print(
        """
    *** Warning: this script is "pre-alpha" software ***

    This scripts reads in `marks.csv` created by `plom-finish`.

    This script then looks for "{}",
    which you should have exported from Canvas.
    This file must have an existing column which will be filled with
    to.  Canvas columns have names like "XXXX (<number>)".
    This script is going to try to complete the column name from
    "{}"

    It outputs a new .csv file for importing back into Canvas:
    "{}"

    Read "docs/returning_papers.md" before using this.
    """.format(
            canvas_fromfile, canvas_test_name, canvas_grades_tofile
        )
    )
    input("Press Enter to continue...")

    print()
    make_canvas_gradefile(
        canvas_fromfile, canvas_grades_tofile, test_parthead=canvas_test_name
    )



#!/usr/bin/env python3

# Copyright (C) 2020 Andrew Rechnitzer
# Copyright (C) 2020-2021 Colin B. Macdonald
# SPDX-License-Identifier: AGPL-3.0-or-later

"""Plom script for post-grading tasks.

## Overview of the "finishing" process

  1. Use the `status` command to check on progress.

  2. Use the `csv` command to produce a CSV file summarizing
     completed papers and marks (so far, if marking is ongoing).

  3. Run the `reassemble` command build PDFs of marked papers.


## Digital return

The reassembled PDF files can be returned to students in various ways.
The `webpage` command builds a webpage with individualized secret codes
to be distributed to each student e.g., via Canvas or another LMS.
"""

__copyright__ = "Copyright (C) 2020-2021 Andrew Rechnitzer, Colin B. Macdonald et al"
__credits__ = "The Plom Project Developers"
__license__ = "AGPL-3.0-or-later"

import argparse
import os
from textwrap import dedent

from plom import __version__

# TODO: be more decisive about how this should be
from plom.finish import clear_manager_login
import plom.finish.check_completed
import plom.finish.spreadsheet
from plom.finish.spreadsheet import CSVFilename
import plom.finish.reassemble_completed
import plom.finish.reassemble_ID_only
import plom.finish.coded_return


parser = argparse.ArgumentParser(
    description=__doc__.split("\n")[0],
    epilog="\n".join(__doc__.split("\n")[1:]),
    formatter_class=argparse.RawDescriptionHelpFormatter,
)
parser.add_argument("--version", action="version", version="%(prog)s " + __version__)

sub = parser.add_subparsers(dest="command")

spCheck = sub.add_parser(
    "status",
    help="How's progress?",
    description="List progress and which tests that have been completed.",
)
spCSV = sub.add_parser(
    "csv",
    help="CSV file with marks/progress info",
    description='Create a spreadsheet of grades named "{}".'.format(CSVFilename),
    epilog="""
        If grading is not yet complete, the spreadsheet contains partial info
        and any warnings so far.
    """,
)
spAssemble = sub.add_parser(
    "reassemble",
    help="Create PDFs to return to students",
    description="""
        After papers have been ID'd and marked, this command builds PDFs
        to return to students.  A special case deals with the online-return
        of papers that were marked offline (before scanning).
    """,
    epilog="""
        WARNING: This command must be run on the server, and in the
        server's directory (where you ran `plom-server launch`).
        This may change in the future.
    """,
)
spAssemble.add_argument(
    "--ided_only",
    action="store_true",
    help="Reassemble PDF files for ID'ed (but offline-graded) papers.",
)
spCodedReturn = sub.add_parser(
    "webpage",
    help="Create HTML page for digital return",
    description="Prepare HTML page for return using out-of-band per-student secret codes.",
    epilog=dedent(
        """
        The webpage will be in `codedReturn` and the secret codes in
        `return_codes.csv`.

        There may be scripts in `share/plom/contrib` to assist with
        distributing the secret codes.

        This command must have access to the results of `reassemble`.
    """
    ),
    formatter_class=argparse.RawDescriptionHelpFormatter,
)
spCodedReturn.add_argument(
    "--hex",
    action="store_true",
    help="""
        Use a string of hexadecimal instead of decimal digits for the
        secret codes.
        More secure but may cause problems if you use certain Canvas
        workarounds to distribute the codes.
    """,
)
spCodedReturn.add_argument(
    "--digits",
    type=int,
    default=9,
    metavar="N",
    action="store",
    help="Length of the secret code.  Defaults to 9.",
)
spCodedReturn.add_argument(
    "--salt",
    type=str,
    help="""
        Instead of random codes, use a hash of the student ID, salted
        with the string SALT.  The codes will then be reproducible by
        anyone who knows this string (and the student IDs).
        As its susceptible to offline attacks, a longer string is
        recommended: you can put quotes around a phrase e.g.,
        `--salt "Many a slip twixt the cup and the lip"`.
    """,
)
spClear = sub.add_parser(
    "clear",
    help='Clear "manager" login',
    description='Clear "manager" login after a crash or other expected event.',
)
for x in (spCheck, spCSV, spAssemble, spClear):
    x.add_argument("-s", "--server", metavar="SERVER[:PORT]", action="store")
    x.add_argument("-w", "--password", type=str, help='for the "manager" user')


def main():
    args = parser.parse_args()

    if not hasattr(args, "server") or not args.server:
        try:
            args.server = os.environ["PLOM_SERVER"]
        except KeyError:
            pass
    if not hasattr(args, "password") or not args.password:
        try:
            args.password = os.environ["PLOM_MANAGER_PASSWORD"]
        except KeyError:
            pass

    if args.command == "status":
        plom.finish.check_completed.main(args.server, args.password)
    elif args.command == "csv":
        plom.finish.spreadsheet.main(args.server, args.password)
    elif args.command == "reassemble":
        if args.ided_only:
            plom.finish.reassemble_ID_only.main(args.server, args.password)
        else:
            plom.finish.reassemble_completed.main(args.server, args.password)
    elif args.command == "webpage":
        plom.finish.coded_return.main(args.hex, args.digits, args.salt)
    elif args.command == "clear":
        clear_manager_login(args.server, args.password)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



__author__ = "Andrew Rechnitzer"
__copyright__ = "Copyright (C) 2018-2020 Andrew Rechnitzer"
__credits__ = ["Andrew Rechnitzer", "Colin Macdonald", "Elvis Cai", "Matt Coles"]
__license__ = "AGPLv3"

from PyQt5.QtCore import Qt, pyqtSignal, QSize
from PyQt5.QtGui import QIcon, QPixmap
from PyQt5.QtWidgets import (
    QAbstractItemView,
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QFormLayout,
    QFrame,
    QLabel,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QTableView,
    QToolButton,
    QVBoxLayout,
)
from plom import isValidStudentNumber


class ErrorMessage(QMessageBox):
    """A simple error message pop-up"""

    def __init__(self, txt):
        super().__init__()
        self.setText(txt)
        self.setStandardButtons(QMessageBox.Ok)


class SimpleMessage(QMessageBox):
    """A simple message pop-up with yes/no buttons."""

    def __init__(self, txt):
        super().__init__()
        self.setText(txt)
        self.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        self.setDefaultButton(QMessageBox.Yes)


class SimpleMessageCheckBox(QMessageBox):
    """A simple message pop-up with yes/no buttons and a checkbox.

    Args:
        txt: plaintext or html content for the dialog
        cbtxt: optional text for the checkbox else default
    """

    def __init__(self, txt, cbtxt=None):
        super().__init__()
        if cbtxt:
            self.cb = QCheckBox(cbtxt)
        else:
            self.cb = QCheckBox("Don't show this message again")
        self.setText(txt)
        self.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        self.setDefaultButton(QMessageBox.Yes)
        self.setCheckBox(self.cb)


class SimpleTableView(QTableView):
    """A table-view widget that emits annotateSignal when
    the user hits enter or return.
    """

    # This is picked up by the marker, lets it know to annotate
    annotateSignal = pyqtSignal()

    def __init__(self, parent=None):
        super(SimpleTableView, self).__init__()
        # User can sort, cannot edit, selects by rows.
        self.setSortingEnabled(True)
        self.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        # Resize to fit the contents
        self.resizeRowsToContents()
        self.horizontalHeader().setStretchLastSection(True)

    def keyPressEvent(self, event):
        # If user hits enter or return, then fire off
        # the annotateSignal, else pass the event on.
        key = event.key()
        if key == Qt.Key_Return or key == Qt.Key_Enter:
            self.annotateSignal.emit()
        else:
            super(SimpleTableView, self).keyPressEvent(event)


class SimpleToolButton(QToolButton):
    """Specialise the tool button to be an icon above text."""

    def __init__(self, txt, icon):
        super(SimpleToolButton, self).__init__()
        self.setText(txt)
        self.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
        self.setIcon(QIcon(QPixmap(icon)))
        self.setIconSize(QSize(24, 24))
        self.setMinimumWidth(100)


class NoAnswerBox(QDialog):
    def __init__(self):
        super(NoAnswerBox, self).__init__()
        self.setWindowTitle("Is this answer blank?")
        self.yesNextB = QPushButton("Yes and &Next")
        self.yesDoneB = QPushButton("&Yes")
        self.noB = QPushButton("No, &cancel")
        self.yesNextB.clicked.connect(lambda: self.done(1))
        self.yesDoneB.clicked.connect(lambda: self.done(2))
        self.noB.clicked.connect(self.reject)
        grid = QGridLayout()
        moreinfo = QLabel(
            "<p>Is this answer blank or nearly blank?</p>"
            "<p>Please answer &ldquo;no&rdquo; if there is "
            "<em>any possibility</em> of relevant writing on the page.</p>"
        )
        moreinfo.setWordWrap(True)
        grid.addWidget(moreinfo, 0, 1, 1, 2)
        grid.addWidget(QLabel("advance to next paper"), 1, 2)
        grid.addWidget(QLabel("keep annotating"), 2, 2)
        grid.addWidget(QLabel("keep annotating"), 3, 2)
        grid.addWidget(self.yesNextB, 1, 1)
        grid.addWidget(self.yesDoneB, 2, 1)
        grid.addWidget(self.noB, 3, 1)
        self.setLayout(grid)


class BlankIDBox(QDialog):
    def __init__(self, parent, testNumber):
        super(BlankIDBox, self).__init__()
        self.parent = parent
        self.testNumber = testNumber
        self.setWindowTitle("What is blank on test/paper {}?".format(testNumber))
        grid = QGridLayout()

        grid.addWidget(
            QLabel("Please scan through the whole paper before continuing."), 0, 1, 1, 2
        )

        self.blankB = QPushButton("Whole paper is &blank")
        self.noIDB = QPushButton("&No ID given but not blank")
        self.noB = QPushButton("&Cancel")

        self.blankB.clicked.connect(lambda: self.done(1))
        self.noIDB.clicked.connect(lambda: self.done(2))
        self.noB.clicked.connect(self.reject)
        grid.addWidget(QLabel("Please check to confirm!"), 1, 2)
        grid.addWidget(
            QLabel("There is writing on other this or other pages."),
            2,
            2,
        )
        grid.addWidget(self.blankB, 1, 1)
        grid.addWidget(self.noIDB, 2, 1)
        grid.addWidget(self.noB, 3, 1)
        self.setLayout(grid)


class SNIDBox(QDialog):
    def __init__(self, id_name_text):
        super(SNIDBox, self).__init__()
        self.sidLE = QLineEdit()
        self.snameLE = QLineEdit()
        self.guessInput(id_name_text)
        self.okB = QPushButton("&Done")
        self.cancelB = QPushButton("&Cancel")
        fl = QFormLayout()
        fl.addRow(QLabel("Student ID:"), self.sidLE)
        fl.addRow(QLabel("Student name:"), self.snameLE)
        fl.addRow(self.okB)
        fl.addRow(self.cancelB)
        self.setLayout(fl)

        self.okB.clicked.connect(self.check)
        self.cancelB.clicked.connect(self.reject)
        self.sid = ""
        self.sname = ""

    def guessInput(self, id_name_text):
        """Extract the digits from id_name_text and use it to fill the sid-entry, and then extract alphabetic from id_name_text and use it to fill the sname-entry"""
        sid = ""
        sname = ""
        for c in id_name_text:
            # if it is a number add it to sid
            if c.isdigit():
                sid += c
            # if it is alphabetic add it to sname
            elif c.isalpha() or c in [" ", ","]:
                sname += c
            else:
                pass
        self.sidLE.setText(sid.strip())
        self.snameLE.setText(sname.strip())

    def check(self):
        self.sid = self.sidLE.text().strip()
        self.sname = self.snameLE.text().strip()
        if not isValidStudentNumber(self.sid):
            ErrorMessage("Not a valid student number.").exec_()
            return
        if not self.sname:
            ErrorMessage(
                "<p>Student name should not be blank.</p>"
                "<p>(If you cannot read it, use &ldquo;Unknown&rdquo;.)</p>"
            ).exec_()
            return
        self.accept()


class ClientSettingsDialog(QDialog):
    def __init__(self, s):
        super(QDialog, self).__init__()
        # self.parent = parent
        self.setWindowTitle("Plom client options")

        flay = QFormLayout()

        self.comboLog = QComboBox()
        self.comboLog.addItems(["Debug", "Info", "Warning", "Error", "Critical"])
        self.comboLog.setCurrentText(s.get("LogLevel", "Info"))
        flay.addRow("Logging level:", self.comboLog)
        moreinfo = QLabel(
            "(In order of severity; less serious messages will not be logged)"
        )
        flay.addWidget(moreinfo)

        self.checkLogFile = QCheckBox("Log to file (requires restart)")
        self.checkLogFile.setCheckState(
            Qt.Checked if s.get("LogToFile") else Qt.Unchecked
        )
        flay.addWidget(self.checkLogFile)

        line = QFrame()
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        flay.addRow(line)

        self.checkFore = QCheckBox("Force foreground upload/downloads")
        self.checkFore.setCheckState(
            Qt.Checked if s.get("FOREGROUND") else Qt.Unchecked
        )
        flay.addWidget(self.checkFore)

        moreinfo = QLabel(
            "By default, Plom does these operations in background threads.\n"
            "Checking this (e.g., for debugging or paranoia) will result in\n"
            "delays between papers."
        )
        # moreinfo.setWordWrap(True)
        flay.addWidget(moreinfo)

        line = QFrame()
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        flay.addRow(line)

        self.leftHandMouse = QCheckBox("Left-handed mouse")
        self.leftHandMouse.setCheckState(
            Qt.Checked if s.get("mouse").lower() == "left" else Qt.Unchecked
        )
        flay.addWidget(self.leftHandMouse)

        self.checkSidebarOnRight = QCheckBox("Annotator sidebar on right")
        self.checkSidebarOnRight.setCheckState(
            Qt.Checked if s.get("SidebarOnRight") else Qt.Unchecked
        )
        flay.addWidget(self.checkSidebarOnRight)

        line = QFrame()
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        flay.addRow(line)

        self.checkWarnCom = QCheckBox(
            "Warn on insufficient feedback (e.g., no comments)"
        )
        self.checkWarnMark = QCheckBox("Warn if score is inconsistent with annotations")
        flay.addWidget(self.checkWarnCom)
        flay.addWidget(self.checkWarnMark)
        self.checkWarnCom.setCheckState(
            Qt.Checked if s.get("CommentsWarnings") else Qt.Unchecked
        )
        self.checkWarnMark.setCheckState(
            Qt.Checked if s.get("MarkWarnings") else Qt.Unchecked
        )
        if not s.get("POWERUSER"):
            self.checkWarnCom.setEnabled(False)
            self.checkWarnMark.setEnabled(False)

        buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)

        vlay = QVBoxLayout()
        vlay.addLayout(flay)
        vlay.addWidget(buttons)
        self.setLayout(vlay)

        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)

    def getStuff(self):
        return (
            self.checkFore.checkState() == Qt.Checked,
            self.comboLog.currentText(),
            self.checkLogFile.checkState() == Qt.Checked,
            self.checkWarnCom.checkState() == Qt.Checked,
            self.checkWarnMark.checkState() == Qt.Checked,
            self.leftHandMouse.checkState() == Qt.Checked,
            self.checkSidebarOnRight.checkState() == Qt.Checked,
        )



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2019 Andrew Rechnitzer
# Copyright (C) 2018 Elvis Cai
# Copyright (C) 2019-2020 Colin B. Macdonald

import logging

from PyQt5.QtWidgets import (
    QWidget,
    QPushButton,
    QToolButton,
    QButtonGroup,
    QGridLayout,
    QMenu,
    QSizePolicy,
)


log = logging.getLogger("markhndlr")


class MarkHandler(QWidget):
    def __init__(self, parent, maxScore, markStyle):
        super(MarkHandler, self).__init__()
        self.parent = parent
        # Set max score/mark
        self.maxScore = maxScore
        # Set current score/mark.
        self.currentScore = 0
        # One button for each possible mark, and a dictionary to store them.
        self.markButtons = {}
        # Styling for buttons
        self.redStyle = (
            "border: 2px solid #ff0000; background: "
            "qlineargradient(x1:0,y1:0,x2:1,y2:0, stop: 0 #ff0000, "
            "stop: 0.3 #ffcccc, stop: 0.7 #ffcccc, stop: 1 #ff0000);"
        )
        # By default we set style to marking-UP.
        self.style = "Up"
        # Keep last delta used
        self.lastDelta = 0
        self.setLayout(QGridLayout())
        self._setStyle(markStyle)
        self.setDeltaButtonMenu()

    def _setStyle(self, markStyle):
        """Sets the mark entry style - either total, up or down
        Total - user just clicks the total mark.
        Up - score starts at zero and increments.
        Down - score starts at max and decrements.
        """
        # if passed a marking style, then set up accordingly.
        if markStyle == 1:
            self.setMarkingTotal()
        elif markStyle == 3:
            self.setMarkingDown()
        else:
            # Default to mark-up.
            self.setMarkingUp()
        self.ve = QButtonGroup()
        self.ve.setExclusive(True)
        for s, x in self.markButtons.items():
            self.ve.addButton(x)

    def resetAndMaybeChange(self, maxScore, markStyle):
        """Reset score, replace max/style with new values, regen buttons.

        Total - user just clicks the total mark.
        Up - score starts at zero and increments.
        Down - score starts at max and decrements.
        """
        d = {"Total": 1, "Up": 2, "Down": 3}
        if self.maxScore == maxScore and d[self.style] == markStyle:
            if markStyle == 3:
                self.setMark(self.maxScore)
            else:
                self.setMark(0)
            self.updateRelevantDeltaActions()
            return
        log.info("Adjusting for new number or new style")
        for k, x in self.markButtons.items():
            self.ve.removeButton(x)
            self.layout().removeWidget(x)
            x.deleteLater()
        self.markButtons.clear()
        self.currentScore = 0
        self.maxScore = maxScore
        self.markButtons = {}
        self._setStyle(markStyle)
        self.setDeltaButtonMenu()

    def getButtonList(self):
        """Return a list of the buttons"""
        return self.markButtons.values()

    def setMarkingUp(self):
        self.setMark(0)
        grid = self.layout()
        # assume our container will deal with margins
        grid.setContentsMargins(0, 0, 0, 0)
        grid.setSpacing(3)
        ncolumn = min(self.maxScore, 5)

        for k in range(0, self.maxScore + 1):
            b = QToolButton()
            self.markButtons[k] = b
            b.setText("+{}".format(k))
            b.setCheckable(True)
            # b.setAutoExclusive(True)
            grid.addWidget(b, k // ncolumn + 1, k % ncolumn, 1, 1)
            b.clicked.connect(self.setDeltaMark)
            b.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)

        self.style = "Up"

    def setMarkingDown(self):
        self.setMark(self.maxScore)
        grid = self.layout()
        # assume our container will deal with margins
        grid.setContentsMargins(0, 0, 0, 0)
        grid.setSpacing(3)
        ncolumn = min(self.maxScore, 5)

        for k in range(0, self.maxScore + 1):
            b = QToolButton()
            self.markButtons[k] = b
            b.setText("-{}".format(k))
            b.setCheckable(True)
            # b.setAutoExclusive(True)
            grid.addWidget(b, k // ncolumn + 1, k % ncolumn, 1, 1)
            b.clicked.connect(self.setDeltaMark)
            b.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)

        self.parent.totalMarkSet(self.currentScore)
        self.style = "Down"

    def setMarkingTotal(self):
        grid = self.layout()
        self.ptmb = QPushButton()

        if self.maxScore > 5:
            ncolumn = 3
        else:
            ncolumn = 2

        for k in range(0, self.maxScore + 1):
            self.markButtons[k] = QPushButton("{}".format(k))
            grid.addWidget(self.markButtons[k], k // ncolumn, k % ncolumn)
            self.markButtons[k].clicked.connect(self.setTotalMark)
            self.markButtons[k].setSizePolicy(
                QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding
            )

        self.parent.totalMarkSet(self.currentScore)
        self.style = "Total"

    def setDeltaMark(self):
        self.sender().setChecked(True)
        # yuck, but this will ensure other tools are not checked
        self.parent.ui.deltaButton.setChecked(True)
        self.currentDelta = self.sender().text()
        self.parent.deltaMarkSet(self.currentDelta)

    def setTotalMark(self):
        self.ptmb.setStyleSheet("")
        self.ptmb = self.sender()
        self.ptmb.setStyleSheet(self.redStyle)
        self.currentScore = int(self.sender().text())
        self.parent.totalMarkSet(self.currentScore)

    def setMark(self, newScore):
        self.currentScore = newScore
        self.parent.totalMarkSet(self.currentScore)

    def clearButtonStyle(self):
        # TODO: kill this?  but how to uncheck all in an exclusive group?
        if self.style == "Total":
            pass  # don't clear the styling when marking total.
        else:
            # Try this one wierd trick...
            self.ve.setExclusive(False)
            for k, x in self.markButtons.items():
                x.setChecked(False)
            self.ve.setExclusive(True)

    def loadDeltaValue(self, delta):
        # delta is a string
        idelta = int(delta)
        if abs(idelta) > self.maxScore or self.style == "Total":
            return
        self.markButtons[abs(idelta)].animateClick()

    def unpickleTotal(self, score):
        if (score <= self.maxScore) and (score >= 0) and (self.style == "Total"):
            self.markButtons[score].animateClick()

    def incrementDelta(self, dlt):
        # dlt is a string, so make int(delta)
        delta = int(dlt)
        if self.style == "Up":
            if delta < 0:
                delta = 0
            else:
                delta += 1
            if delta > self.maxScore:
                delta = 0
            self.markButtons[delta].animateClick()
        elif self.style == "Down":
            if delta > 0:
                delta = -1
            else:
                delta -= 1
            if abs(delta) > self.maxScore:
                delta = -1
            self.markButtons[-delta].animateClick()

    def clickDelta(self, dlt):
        # dlt is a string, so make int(delta)
        # be careful if this has been set by a no-delta comment.
        if dlt == ".":
            delta = 0
        else:
            delta = int(dlt)
        if self.style == "Up":
            if delta < 0:
                delta = 0
            self.markButtons[delta].animateClick()
        elif self.style == "Down":
            if delta > 0:
                delta = 0
            self.markButtons[-delta].animateClick()

    def setDeltaButtonMenu(self):
        if self.style == "Total":
            # mark total - don't set anything
            return
        deltaMenu = QMenu("Set Delta")
        self.deltaActions = {}
        if self.style == "Up":
            # set to mark up
            for k in range(0, self.maxScore + 1):
                self.deltaActions[k] = deltaMenu.addAction("+{}".format(k))
                self.deltaActions[k].triggered.connect(self.markButtons[k].animateClick)
        elif self.style == "Down":
            # set to mark down
            for k in range(0, self.maxScore + 1):
                self.deltaActions[k] = deltaMenu.addAction("-{}".format(k))
                self.deltaActions[k].triggered.connect(self.markButtons[k].animateClick)
        self.parent.ui.deltaButton.setMenu(deltaMenu)
        self.updateRelevantDeltaActions()

    def updateRelevantDeltaActions(self):
        if self.style == "Total":
            return
        elif self.style == "Up":
            for k in range(0, self.maxScore + 1):
                if self.currentScore + k <= self.maxScore:
                    self.deltaActions[k].setEnabled(True)
                    self.markButtons[k].setEnabled(True)
                else:
                    self.deltaActions[k].setEnabled(False)
                    self.markButtons[k].setEnabled(False)
        elif self.style == "Down":
            for k in range(0, self.maxScore + 1):
                if self.currentScore >= k:
                    self.deltaActions[k].setEnabled(True)
                    self.markButtons[k].setEnabled(True)
                else:
                    self.deltaActions[k].setEnabled(False)
                    self.markButtons[k].setEnabled(False)



__author__ = "Andrew Rechnitzer"
__copyright__ = "Copyright (C) 2020 Andrew Rechnitzer"
__credits__ = ["Andrew Rechnitzer", "Colin Macdonald", "Elvis Cai", "Matt Coles"]
__license__ = "AGPLv3"

from PyQt5.QtCore import Qt, QStringListModel
from PyQt5.QtGui import QBrush, QGuiApplication, QPainter, QPixmap
from PyQt5.QtWidgets import (
    QCompleter,
    QDialog,
    QFrame,
    QFormLayout,
    QGraphicsPixmapItem,
    QGraphicsItemGroup,
    QGraphicsScene,
    QGraphicsView,
    QGridLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSpinBox,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)


class ActionTab(QWidget):
    def __init__(self, parent):
        super(ActionTab, self).__init__()
        self.parent = parent
        vb = QVBoxLayout()
        self.db = QPushButton("Discard Page")
        self.eb = QPushButton("Extra Page")
        self.tb = QPushButton("Test Page")
        self.hb = QPushButton("Homework Page")
        vb.addWidget(self.eb)
        vb.addWidget(self.tb)
        vb.addWidget(self.hb)
        vb.addWidget(self.db)
        self.setLayout(vb)
        self.show()
        self.db.clicked.connect(self.discard)
        self.hb.clicked.connect(self.homework)
        self.eb.clicked.connect(self.extra)
        self.tb.clicked.connect(self.test)

    def discard(self):
        self.parent.optionTW.setCurrentIndex(4)

    def extra(self):
        self.parent.optionTW.setCurrentIndex(1)

    def test(self):
        self.parent.optionTW.setCurrentIndex(2)

    def homework(self):
        self.parent.optionTW.setCurrentIndex(3)


class DiscardTab(QWidget):
    def __init__(self, parent):
        super(DiscardTab, self).__init__()
        self.parent = parent
        vb = QVBoxLayout()
        self.db = QPushButton("Click to confirm discard")
        self.ob = QPushButton("Return to other options")
        vb.addStretch(0)
        vb.addWidget(self.db)
        vb.addStretch(0)
        vb.addWidget(self.ob)
        self.setLayout(vb)
        self.show()
        self.db.clicked.connect(self.discard)
        self.ob.clicked.connect(self.other)

    def discard(self):
        self.parent.action = "discard"
        self.parent.accept()

    def other(self):
        self.parent.optionTW.setCurrentIndex(0)


class ExtraTab(QWidget):
    def __init__(self, parent, maxT, maxQ):
        super(ExtraTab, self).__init__()
        self.parent = parent
        vb = QVBoxLayout()
        fl = QFormLayout()
        self.frm = QFrame()
        self.ob = QPushButton("Return to other options")
        self.tsb = QSpinBox()
        self.qsb = QSpinBox()
        self.tsb.setMinimum(1)
        self.tsb.setMaximum(maxT)
        self.qsb.setMinimum(1)
        self.qsb.setMaximum(maxQ)
        self.cb = QPushButton("Click to confirm")
        self.vqb = QPushButton("View that question")
        self.vwb = QPushButton("View whole test")
        fl.addRow(QLabel("Test number:"), self.tsb)
        fl.addRow(QLabel("Question number:"), self.qsb)
        fl.addRow(self.vqb)
        fl.addRow(self.vwb)
        fl.addRow(self.cb)
        self.frm.setLayout(fl)
        vb.addWidget(self.frm)
        vb.addStretch(1)
        vb.addWidget(self.ob)
        self.setLayout(vb)
        self.show()
        self.vqb.clicked.connect(self.viewQuestion)
        self.vwb.clicked.connect(self.viewWholeTest)
        self.cb.clicked.connect(self.confirm)
        self.ob.clicked.connect(self.other)

    def confirm(self):
        self.parent.action = "extra"
        self.parent.test = self.tsb.value()
        self.parent.pq = self.qsb.value()
        self.parent.accept()

    def viewQuestion(self):
        self.parent.viewQuestion(self.tsb.value(), self.qsb.value())

    def viewWholeTest(self):
        self.parent.viewWholeTest(self.tsb.value())

    def other(self):
        self.parent.optionTW.setCurrentIndex(0)


class HWTab(QWidget):
    def __init__(self, parent, maxQ, iDict):
        super(HWTab, self).__init__()
        self.parent = parent
        vb = QVBoxLayout()
        fl = QFormLayout()
        self.frm = QFrame()
        self.ob = QPushButton("Return to other options")
        self.sidle = QLineEdit()
        # set up sid completion
        self.sidTestDict = {"{}: {}".format(iDict[x][0], iDict[x][1]): x for x in iDict}
        self.sidlist = QStringListModel()
        self.sidlist.setStringList([x for x in self.sidTestDict])
        self.sidcompleter = QCompleter()
        self.sidcompleter.setCaseSensitivity(Qt.CaseInsensitive)
        self.sidcompleter.setFilterMode(Qt.MatchContains)
        self.sidcompleter.setModel(self.sidlist)
        self.sidle.setCompleter(self.sidcompleter)
        # now set up other gui elements
        self.qsb = QSpinBox()
        self.qsb.setMinimum(1)
        self.qsb.setMaximum(maxQ)
        self.testl = QLabel("")
        self.cb = QPushButton("Click to confirm")
        self.vqb = QPushButton("View that question")
        self.vwb = QPushButton("View whole test")
        fl.addRow(QLabel("Student ID / Name:"), self.sidle)
        fl.addRow(QLabel("Question number:"), self.qsb)
        fl.addRow(QLabel("Test number:"), self.testl)
        fl.addRow(self.vqb)
        fl.addRow(self.vwb)
        fl.addRow(self.cb)
        self.frm.setLayout(fl)
        vb.addWidget(self.frm)
        vb.addStretch(1)
        vb.addWidget(self.ob)
        self.setLayout(vb)
        self.show()
        self.vqb.clicked.connect(self.viewQuestion)
        self.vwb.clicked.connect(self.viewWholeTest)
        self.cb.clicked.connect(self.confirm)
        self.ob.clicked.connect(self.other)
        self.sidle.returnPressed.connect(self.checkID)
        # check ID when user clicks on entry in completer pop-up - not just when return pressed
        self.sidcompleter.activated.connect(self.checkID)

    def checkID(self):
        sid = self.sidle.text()
        if sid in self.sidTestDict:
            self.testl.setText(self.sidTestDict[sid])
        else:
            self.testl.setText("")

    def confirm(self):
        if self.testl.text() == "":
            return
        self.parent.action = "homework"
        self.parent.sid = self.sidle.text()
        self.parent.pq = self.qsb.value()
        self.parent.test = int(self.testl.text())
        self.parent.accept()

    def viewQuestion(self):
        if self.testl.text() == "":
            return
        else:
            self.parent.viewQuestion(int(self.testl.text()), self.qsb.value())

    def viewWholeTest(self):
        if self.testl.text() == "":
            return
        else:
            self.parent.viewWholeTest(int(self.testl.text()))

    def other(self):
        self.parent.optionTW.setCurrentIndex(0)


class TestTab(QWidget):
    def __init__(self, parent, maxT, maxP):
        super(TestTab, self).__init__()
        self.parent = parent
        vb = QVBoxLayout()
        fl = QFormLayout()
        self.frm = QFrame()
        self.ob = QPushButton("Return to other options")
        self.tsb = QSpinBox()
        self.psb = QSpinBox()
        self.tsb.setMinimum(1)
        self.tsb.setMaximum(maxT)
        self.psb.setMinimum(1)
        self.psb.setMaximum(maxP)
        self.cb = QPushButton("Click to confirm")
        self.cpb = QPushButton("Check that page")
        self.vwb = QPushButton("View whole test")
        fl.addRow(QLabel("Test number:"), self.tsb)
        fl.addRow(QLabel("Page number:"), self.psb)
        fl.addRow(self.cpb)
        fl.addRow(self.vwb)
        fl.addRow(self.cb)
        self.frm.setLayout(fl)
        vb.addWidget(self.frm)
        vb.addStretch(1)
        vb.addWidget(self.ob)
        self.setLayout(vb)
        self.show()
        self.cpb.clicked.connect(self.checkTPage)
        self.vwb.clicked.connect(self.viewWholeTest)
        self.cb.clicked.connect(self.confirm)
        self.ob.clicked.connect(self.other)

    def confirm(self):
        self.parent.action = "test"
        self.parent.test = self.tsb.value()
        self.parent.pq = self.psb.value()
        self.parent.accept()

    def checkTPage(self):
        self.parent.checkTPage(self.tsb.value(), self.psb.value())

    def viewWholeTest(self):
        self.parent.viewWholeTest(self.tsb.value())

    def other(self):
        self.parent.optionTW.setCurrentIndex(0)


class UnknownView(QGraphicsView):
    """Simple extension of QGraphicsView
    - containing an image and click-to-zoom/unzoom
    """

    def __init__(self, fnames):
        QGraphicsView.__init__(self)
        self.initUI(fnames)

    def initUI(self, fnames):
        # Make QGraphicsScene
        self.scene = QGraphicsScene()
        # TODO = handle different image sizes.
        self.images = {}
        self.imageGItem = QGraphicsItemGroup()
        self.scene.addItem(self.imageGItem)
        self.updateImage(fnames)
        self.setBackgroundBrush(QBrush(Qt.darkCyan))

    def updateImage(self, fnames):
        """Update the image with that from filename"""
        for n in self.images:
            self.imageGItem.removeFromGroup(self.images[n])
            self.images[n].setVisible(False)
        if fnames is not None:
            x = 0
            n = 0
            for fn in fnames:
                self.images[n] = QGraphicsPixmapItem(QPixmap(fn))
                self.images[n].setTransformationMode(Qt.SmoothTransformation)
                self.images[n].setPos(x, 0)
                self.images[n].setVisible(True)
                self.scene.addItem(self.images[n])
                x += self.images[n].boundingRect().width() + 10
                self.imageGItem.addToGroup(self.images[n])
                n += 1

        # Set sensible sizes and put into the view, and fit view to the image.
        br = self.imageGItem.boundingRect()
        self.scene.setSceneRect(
            0,
            0,
            max(1000, br.width()),
            max(1000, br.height()),
        )
        self.setScene(self.scene)
        self.fitInView(self.imageGItem, Qt.KeepAspectRatio)

    def mouseReleaseEvent(self, event):
        """Left/right click to zoom in and out"""
        if (event.button() == Qt.RightButton) or (
            QGuiApplication.queryKeyboardModifiers() == Qt.ShiftModifier
        ):
            self.scale(0.8, 0.8)
        else:
            self.scale(1.25, 1.25)
        self.centerOn(event.pos())

    def resetView(self):
        """Reset the view to its reasonable initial state."""
        self.fitInView(self.imageGItem, Qt.KeepAspectRatio)

    def rotateImage(self, dTheta):
        self.rotate(dTheta)
        self.resetView()


class UnknownViewWindow(QDialog):
    """Simple view window for pageimages"""

    def __init__(self, parent, fnames, tpq, iDict):
        QWidget.__init__(self)
        self.parent = parent
        self.numberOfTests = tpq[0]
        self.numberOfPages = tpq[1]
        self.numberOfQuestions = tpq[2]
        self.iDict = iDict

        if type(fnames) == list:
            self.initUI(fnames)
        else:
            self.initUI([fnames])
        self.action = ""
        self.test = 0
        self.pq = 0
        self.sid = ""

    def initUI(self, fnames):
        # Grab an UnknownView widget (QGraphicsView)
        self.view = UnknownView(fnames)
        # Render nicely
        self.view.setRenderHint(QPainter.HighQualityAntialiasing)
        self.optionTW = QTabWidget()

        # reset view button passes to the UnknownView.
        self.resetB = QPushButton("reset view")
        self.rotatePlusB = QPushButton("rotate +90")
        self.rotateMinusB = QPushButton("rotate -90")
        self.cancelB = QPushButton("&cancel")
        self.maxNormB = QPushButton("&max/norm")

        self.cancelB.clicked.connect(self.reject)
        self.resetB.clicked.connect(lambda: self.view.resetView())
        self.rotatePlusB.clicked.connect(self.rotatePlus)
        self.rotateMinusB.clicked.connect(self.rotateMinus)
        self.maxNormB.clicked.connect(self.swapMaxNorm)

        self.resetB.setAutoDefault(False)  # return wont click the button by default.
        self.rotatePlusB.setAutoDefault(False)
        self.rotateMinusB.setAutoDefault(False)

        # Layout simply
        grid = QGridLayout()
        grid.addWidget(self.view, 1, 1, 10, 6)
        grid.addWidget(self.optionTW, 2, 17, 8, 4)
        grid.addWidget(self.resetB, 20, 1)
        grid.addWidget(self.rotatePlusB, 20, 2)
        grid.addWidget(self.rotateMinusB, 20, 3)
        grid.addWidget(self.cancelB, 20, 20)
        grid.addWidget(self.maxNormB, 1, 20)
        self.setLayout(grid)
        self.show()
        # Store the current exam view as a qtransform
        self.viewTrans = self.view.transform()
        self.dx = self.view.horizontalScrollBar().value()
        self.dy = self.view.verticalScrollBar().value()
        self.theta = 0
        self.initTabs()

    def updateImage(self, fnames):
        """Pass file to the view to update the image"""
        # first store the current view transform and scroll values
        self.viewTrans = self.view.transform()
        self.dx = self.view.horizontalScrollBar().value()
        self.dy = self.view.verticalScrollBar().value()
        # update the image
        if type(fnames) == list:
            self.view.updateImage(fnames)
        else:
            self.view.updateImage([fnames])

        # re-set the view transform and scroll values
        self.view.setTransform(self.viewTrans)
        self.view.horizontalScrollBar().setValue(self.dx)
        self.view.verticalScrollBar().setValue(self.dy)

    def initTabs(self):
        self.t0 = ActionTab(self)
        self.t1 = ExtraTab(self, self.numberOfTests, self.numberOfQuestions)
        self.t2 = TestTab(self, self.numberOfTests, self.numberOfPages)
        self.t3 = HWTab(self, self.numberOfQuestions, self.iDict)
        self.t4 = DiscardTab(self)
        self.optionTW.addTab(self.t0, "Actions")
        self.optionTW.addTab(self.t1, "Extra Page")
        self.optionTW.addTab(self.t2, "Test Page")
        self.optionTW.addTab(self.t3, "Homework Page")
        self.optionTW.addTab(self.t4, "Discard")

    def rotatePlus(self):
        self.theta += 90
        if self.theta == 360:
            self.theta = 0
        self.view.rotateImage(90)

    def rotateMinus(self):
        self.theta -= 90
        if self.theta == -90:
            self.theta = 270
        self.view.rotateImage(-90)

    def swapMaxNorm(self):
        """Toggles the window size between max and normal"""
        if self.windowState() != Qt.WindowMaximized:
            self.setWindowState(Qt.WindowMaximized)
        else:
            self.setWindowState(Qt.WindowNoState)

    def viewQuestion(self, testNumber, questionNumber):
        self.parent.viewQuestion(testNumber, questionNumber)

    def viewWholeTest(self, testNumber):
        self.parent.viewWholeTest(testNumber)

    def checkTPage(self, testNumber, pageNumber):
        self.parent.checkTPage(testNumber, pageNumber)
